# -*- coding: utf-8 -*-

"""
* Name:         cognitivefactory.interactive_clustering.clustering.spectral
* Description:  Implementation of constrained spectral clustering algorithms.
* Author:       Erwan Schild
* Created:      17/03/2021
* Licence:      CeCILL-C License v1.0 (https://cecill.info/licences.fr.html)
"""

# ==============================================================================
# IMPORTS PYTHON DEPENDENCIES
# ==============================================================================

from typing import Dict, List, Optional, Union  # To type Python code (mypy).

import numpy as np  # To handle float.
from numpy import ndarray  # To handle matrix and vectors.
from scipy.sparse import csr_matrix  # To handle matrix and vectors.
from sklearn.cluster import SpectralClustering  # To use classical spectral clustering.
from sklearn.metrics import pairwise_kernels  # To compute similary.

from cognitivefactory.interactive_clustering.clustering.abstract import (  # To use abstract interface.; To sort clusters after computation.
    AbstractConstrainedClustering,
    rename_clusters_by_order,
)
from cognitivefactory.interactive_clustering.constraints.abstract import (  # To manage constraints.
    AbstractConstraintsManager,
)


# ==============================================================================
# SPECTRAL CONSTRAINED CLUSTERING
# ==============================================================================
class SpectralConstrainedClustering(AbstractConstrainedClustering):
    """
    This class implements the spectral constrained clustering.
    It inherits from `AbstractConstrainedClustering`.

    References:
        - Spectral Clustering: `Ng, A. Y., M. I. Jordan, et Y.Weiss (2002). On Spectral Clustering: Analysis and an algorithm. In T. G. Dietterich, S. Becker, et Z. Ghahramani (Eds.), Advances in Neural Information Processing Systems 14. MIT Press.`
        - Constrained _'SPEC'_ Spectral Clustering: `Kamvar, S. D., D. Klein, et C. D. Manning (2003). Spectral Learning. Proceedings of the international joint conference on artificial intelligence, 561–566.`

    Examples:
        ```python
        # Import.
        from scipy.sparse import csr_matrix
        from cognitivefactory.interactive_clustering.constraints.binary import BinaryConstraintsManager
        from cognitivefactory.interactive_clustering.clustering.spectral import SpectralConstrainedClustering

        # Create an instance of spectral clustering.
        clustering_model = SpectralConstrainedClustering(
            model="SPEC",
            random_seed=1,
        )

        # Define vectors.
        # NB : use cognitivefactory.interactive_clustering.utils to preprocess and vectorize texts.
        vectors = {
            "0": csr_matrix([1.00, 0.00, 0.00, 0.00]),
            "1": csr_matrix([0.95, 0.02, 0.02, 0.01]),
            "2": csr_matrix([0.98, 0.00, 0.02, 0.00]),
            "3": csr_matrix([0.99, 0.00, 0.01, 0.00]),
            "4": csr_matrix([0.60, 0.17, 0.16, 0.07]),
            "5": csr_matrix([0.60, 0.16, 0.17, 0.07]),
            "6": csr_matrix([0.01, 0.01, 0.01, 0.97]),
            "7": csr_matrix([0.00, 0.01, 0.00, 0.99]),
            "8": csr_matrix([0.00, 0.00, 0.00, 1.00]),
        }

        # Define constraints manager.
        constraints_manager = BinaryConstraintsManager(list_of_data_IDs=list(vectors.keys()))
        constraints_manager.add_constraint(data_ID1="0", data_ID2="1", constraint_type="MUST_LINK")
        constraints_manager.add_constraint(data_ID1="2", data_ID2="3", constraint_type="MUST_LINK")
        constraints_manager.add_constraint(data_ID1="4", data_ID2="5", constraint_type="MUST_LINK")
        constraints_manager.add_constraint(data_ID1="7", data_ID2="8", constraint_type="MUST_LINK")
        constraints_manager.add_constraint(data_ID1="0", data_ID2="4", constraint_type="CANNOT_LINK")
        constraints_manager.add_constraint(data_ID1="2", data_ID2="4", constraint_type="CANNOT_LINK")
        constraints_manager.add_constraint(data_ID1="4", data_ID2="7", constraint_type="CANNOT_LINK")

        # Run clustering.
        dict_of_predicted_clusters = clustering_model(
            constraints_manager=constraints_manager,
            vectors=vectors,
            nb_clusters=3,
        )

        # Print results.
        print("Expected results", ";", {"0": 0, "1": 0, "2": 0, "3": 0, "4": 1, "5": 1, "6": 2, "7": 2, "8": 2,})
        print("Computed results", ":", dict_of_predicted_clusters)
        ```
    """

    # ==============================================================================
    # INITIALIZATION
    # ==============================================================================
    def __init__(
        self, model: str = "SPEC", nb_components: Optional[int] = None, random_seed: Optional[int] = None, **kargs
    ) -> None:
        """
        The constructor for Spectral Constrainted Clustering class.

        Args:
            model (str, optional): The spectral clustering model to use. Available spectral models are `"SPEC"` and `"CCSR"`. Defaults to `"SPEC"`.
            nb_components (Optional[int], optional): The number of eigenvectors to compute in the spectral clustering. If `None`, set the number of components to the number of clusters. Defaults to `None`.
            random_seed (Optional[int], optional): The random seed to use to redo the same clustering. Defaults to `None`.
            **kargs (dict): Other parameters that can be used in the instantiation.

        Raises:
            ValueError: if some parameters are incorrectly set.
        """

        # Store `self.model`.
        if model != "SPEC":  # TODO use `not in {"SPEC"}`. # TODO `"CCSR"` to add after correction.
            raise ValueError("The `model` '" + str(model) + "' is not implemented.")
        self.model: str = model

        # Store `self.nb_components`.
        if (nb_components is not None) and (nb_components < 2):
            raise ValueError(
                "The `nb_components` '" + str(nb_components) + "' must be `None` or greater than or equal to 2."
            )
        self.nb_components: Optional[int] = nb_components

        # Store `self.random_seed`.
        self.random_seed: Optional[int] = random_seed

        # Store `self.kargs` for kmeans clustering.
        self.kargs = kargs

        # Initialize `self.dict_of_predicted_clusters`.
        self.dict_of_predicted_clusters: Optional[Dict[str, int]] = None

    # ==============================================================================
    # MAIN - CLUSTER DATA
    # ==============================================================================
    def cluster(
        self,
        constraints_manager: AbstractConstraintsManager,
        vectors: Dict[str, Union[ndarray, csr_matrix]],
        nb_clusters: int,
        verbose: bool = False,
        **kargs,
    ) -> Dict[str, int]:
        """
        The main method used to cluster data with the Spectral model.

        Args:
            constraints_manager (AbstractConstraintsManager): A constraints manager over data IDs that will force clustering to respect some conditions during computation.
            vectors (Dict[str,Union[ndarray,csr_matrix]]): The representation of data vectors. The keys of the dictionary represents the data IDs. This keys have to refer to the list of data IDs managed by the `constraints_manager`. The value of the dictionary represent the vector of each data. Vectors can be dense (`numpy.ndarray`) or sparse (`scipy.sparse.csr_matrix`).
            nb_clusters (int): The number of clusters to compute.  #TODO Set defaults to None with elbow method or other method ?
            verbose (bool, optional): Enable verbose output. Defaults to `False`.
            **kargs (dict): Other parameters that can be used in the clustering.

        Raises:
            ValueError: if `vectors` and `constraints_manager` are incompatible, or if some parameters are incorrectly set.

        Returns:
            Dict[str,int]: A dictionary that contains the predicted cluster for each data ID.
        """

        ###
        ### GET PARAMETERS
        ###

        # Store `self.constraints_manager` and `self.list_of_data_IDs`.
        if not isinstance(constraints_manager, AbstractConstraintsManager):
            raise ValueError("The `constraints_manager` parameter has to be a `AbstractConstraintsManager` type.")
        self.constraints_manager: AbstractConstraintsManager = constraints_manager
        self.list_of_data_IDs: List[str] = self.constraints_manager.get_list_of_managed_data_IDs()

        # Store `self.vectors`.
        if not isinstance(vectors, dict):
            raise ValueError("The `vectors` parameter has to be a `dict` type.")
        self.vectors: Dict[str, Union[ndarray, csr_matrix]] = vectors

        # Store `self.nb_clusters`.
        if nb_clusters < 2:
            raise ValueError("The `nb_clusters` '" + str(nb_clusters) + "' must be greater than or equal to 2.")
        self.nb_clusters: int = nb_clusters

        # Define `self.current_nb_components`.
        self.current_nb_components: int = (
            self.nb_components
            if ((self.nb_components is not None) and (self.nb_clusters < self.nb_components))
            else self.nb_clusters
        )

        # Compute `self.pairwise_similarity_matrix`.
        self.pairwise_similarity_matrix = csr_matrix(
            [
                [
                    pairwise_kernels(X=self.vectors[data_ID1], Y=self.vectors[data_ID2], metric="rbf")[0][0].astype(
                        np.float64
                    )
                    for data_ID2 in self.list_of_data_IDs
                ]
                for data_ID1 in self.list_of_data_IDs
            ]
        )

        ###
        ### RUN SPECTRAL CONSTRAINED CLUSTERING
        ###

        # Initialize `self.dict_of_predicted_clusters`.
        self.dict_of_predicted_clusters = None

        # Case of `"CCSR"` spectral clustering.
        # TODO Don't work.
        ##if self.model == "CCSR":
        ##    self.dict_of_predicted_clusters = self.clustering_spectral_model_CCSR(verbose=verbose)

        # Case of `"SPEC"` spectral clustering.
        ##### DEFAULTS : if self.model=="SPEC":
        self.dict_of_predicted_clusters = self.clustering_spectral_model_SPEC(verbose=verbose)

        ###
        ### RETURN PREDICTED CLUSTERS
        ###

        return self.dict_of_predicted_clusters

    # ==============================================================================
    # IMPLEMENTATION - SPEC SPECTRAL CLUSTERING
    # ==============================================================================
    def clustering_spectral_model_SPEC(
        self,
        verbose: bool = False,
    ) -> Dict[str, int]:
        """
        Implementation of a simple Spectral clustering algorithm, based affinity matrix modifications.

        References :
            - Constrained _'SPEC'_ Spectral Clustering: `Kamvar, S. D., D. Klein, et C. D. Manning (2003). Spectral Learning. Proceedings of the international joint conference on artificial intelligence, 561–566.`

        Args:
            verbose (bool, optional): Enable verbose output. Default is `False`.

        Returns:
            Dict[str,int]: A dictionary that contains the predicted cluster for each data ID.
        """

        ###
        ### MODIFY CONSTRAINTS MATRIX WITH CONSTRAINTS
        ###

        # Modify the similarity over data IDs.
        for ID1, data_ID1 in enumerate(self.list_of_data_IDs):
            for ID2, data_ID2 in enumerate(self.list_of_data_IDs):

                # Symetry is already handled in next instructions.
                if ID1 > ID2:
                    continue

                # For each `"MUST_LINK"` constraint, set the similarity to 1.0.
                if (
                    self.constraints_manager.get_inferred_constraint(
                        data_ID1=data_ID1,
                        data_ID2=data_ID2,
                    )
                    == "MUST_LINK"
                ):
                    self.pairwise_similarity_matrix[ID1, ID2] = 1.0
                    self.pairwise_similarity_matrix[ID2, ID1] = 1.0

                # For each `"CANNOT_LINK"` constraint, set the similarity to 0.0.
                elif (
                    self.constraints_manager.get_inferred_constraint(
                        data_ID1=data_ID1,
                        data_ID2=data_ID2,
                    )
                    == "CANNOT_LINK"
                ):
                    self.pairwise_similarity_matrix[ID1, ID2] = 0.0
                    self.pairwise_similarity_matrix[ID2, ID1] = 0.0

        ###
        ### RUN SPECTRAL CONSTRAINED CLUSTERING
        ###     | Define laplacian matrix
        ###     | Compute eigen vectors
        ###     | Cluster eigen vectors
        ###     | Return labels based on eigen vectors clustering
        ###

        # Initialize spectral clustering model.
        self.clustering_model = SpectralClustering(
            n_clusters=self.nb_clusters,
            # n_components=self.current_nb_components, #TODO Add if `scikit-learn>=0.24.1`
            affinity="precomputed",
            random_state=self.random_seed,
            **self.kargs,
        )

        # Run spectral clustering model.
        self.clustering_model.fit_predict(X=self.pairwise_similarity_matrix)

        # Get prediction of spectral clustering model.
        list_of_clusters: List[int] = self.clustering_model.labels_.tolist()

        # Define the dictionary of predicted clusters.
        predicted_clusters: Dict[str, int] = {
            data_ID: list_of_clusters[ID] for ID, data_ID in enumerate(self.list_of_data_IDs)
        }

        # Rename cluster IDs by order.
        predicted_clusters = rename_clusters_by_order(clusters=predicted_clusters)

        # Return predicted clusters
        return predicted_clusters

    # ==============================================================================
    # IMPLEMENTATION - CCSR SPECTRAL CLUSTERING
    # ==============================================================================
    """ #TODO : do not work. check if 1) wrong implementation ? 2) cvxopt better ?
    def clustering_spectral_model_CCSR(
        self,
        verbose: bool = False,
    ) -> Dict[str, int]:
        \"""
        Implementation of Constrained Clustering with Spectral Regularization algorithm, based on spectral semidefinite programming interpretation.
        - Source : `Li, Z., Liu, J., & Tang, X. (2009). Constrained clustering via spectral regularization. 2009 IEEE Conference on Computer Vision and Pattern Recognition. https://doi.org/10.1109/cvpr.2009.5206852`
        - MATLAB Implementation : https://icube-forge.unistra.fr/lampert/TSCC/-/tree/master/methods/Li09
        - SOLVERS comparison : https://pypi.org/project/PICOS/
        - CVXPY solver : https://www.cvxpy.org/index.html

        Args:
            verbose (bool, optional): Enable verbose output. Defaults to `False`.

        Returns:
            Dict[str,int]: A dictionary that contains the predicted cluster for each data ID.
        \"""

        ###
        ### COMPUTE NORMALIZED LAPLACIAN
        ###

        # Compute the normalized Laplacian.
        normalized_laplacian, diagonal = csgraph_laplacian(
            csgraph=self.pairwise_similarity_matrix,
            normed=True,
            return_diag=True
        )

        ###
        ### COMPUTE EIGENVECTORS OF LAPLACIAN
        ###

        random.seed(self.random_seed)
        v0 = np.random.rand(min(normalized_laplacian.shape))

        # Compute the largest eigen vectors.
        eigen_values, eigen_vectors = eigsh(
            A=normalized_laplacian,
            which="SM",
            k=self.current_nb_components + 1,
            v0=v0,
        )

        # Ignore eigenvector of eigenvalue 0.
        eigen_values = eigen_values[1:]
        eigen_vectors = eigen_vectors.T[1:]

        if verbose:  # pragma: no cover
            print("EIGENVALUES / EIGENVECTORS")

            for k in range(len(eigen_values)):
                print("    ", "=============")
                print("    ", "ID :         ", k)
                print("    ", "VAL :        ", eigen_values[k])
                print("    ", "VEC :        ", eigen_vectors[k])

        ###
        ### FORMALIZE SEMIDEFINITE PROBLEM
        ###

        ## Problem ::
        ## Cost function to minimize : L(z) = 1/2 * z.T * B * z + b.T * z + c
        ## z : variable to find with the SDP problem
        ## z = vec(M)
        ## M >> 0 (semi definite positive), i.e. M = M.T, M.shape=(nb_components, nb_components)
        ## B = sum ( s_ij s_ij.T )
        # for ij in MUST_LINK or i,j in CANNOT_LINK
        ## b = -2 * sum ( t_ij * s_ij )
        # for ij in MUST_LINK or i,j in CANNOT_LINK
        ## c = sum ( t_ij^2 )
        # for ij in MUST_LINK or i,j in CANNOT_LINK
        ## s_ij = vec( eigen_vector_i.T * eigen_vector_j )
        ## t_ij = 1 if MUST_LINK(i,j), 0 if CANNOT_LINK(i,j)

        # Initialization of SDP variables.
        B = np.zeros((self.current_nb_components ** 2, self.current_nb_components ** 2))
        b = np.zeros((self.current_nb_components ** 2, 1))
        c = 0

        for ID1, data_ID1 in enumerate(self.list_of_data_IDs):
            for ID2, data_ID2 in enumerate(self.list_of_data_IDs):

                # Get eigenvectors.
                eigen_vector_i = np.atleast_2d(eigen_vectors.T[ID1])
                eigen_vector_j = np.atleast_2d(eigen_vectors.T[ID2])

                # Compute eigenvectors similarity.
                U = eigen_vector_j.T @ eigen_vector_i
                s = np.atleast_2d(U.ravel())


                # For each `"MUST_LINK"` constraint, ....
                if self.constraints_manager.get_inferred_constraint(
                    data_ID1=data_ID1,
                    data_ID2=data_ID2,
                ) == "MUST_LINK":

                    # Add the value to SDP variables.
                    B += s.T * s
                    b += - 1 * s.T
                    c += 1 * 1

                # For each `"CANNOT_LINK"` constraint, ....
                if self.constraints_manager.get_inferred_constraint(
                    data_ID1=data_ID1,
                    data_ID2=data_ID2,
                ) == "CANNOT_LINK":

                    # Add the value to SDP variables.
                    B += s.T * s
                    b += - 0 * s.T
                    c += 0 * 0

        ###
        ### SOLVE SEMIDEFINITE PROBLEM
        ###

        # Create a symetric matrix variable.
        M = cp.Variable((self.current_nb_components, self.current_nb_components))

        ### Define constraints.
        SDP_constraints = []

        # The solution must be positive semidefinite.
        SDP_constraints += [M >> 0]

        # Define cost function to minimize : `( 1/2 * z.T * B * z + b.T * z + c )`.
        self.SDP_problem = cp.Problem(
            cp.Minimize(
                cp.quad_form(cp.atoms.affine.vec.vec(M), B)  # `1/2 * z.T * B * z`.
                + b.T @ cp.atoms.affine.vec.vec(M)  # `b.T * z`.
                + c  # c
            ),
            SDP_constraints,
        )

        # Solve the SDP problem.
        self.SDP_problem.solve(solver="MOSEK")
        if verbose:  # pragma: no cover
            print("SEMIDEFINITE PROBLEM")
            print("    ", "STATUS", ":", self.SDP_problem.status)
            print("    ", "COST FUNCTION VALUE", ":", self.SDP_problem.value)

        ###
        ### CLUSTER EIGEN VECTORS
        ###

        # Define square root of M, and force sqrtM to be symetric.
        sqrtM = sqrtm(M.value).real
        sqrtM = (sqrtM + sqrtM.T) / 2

        # Compute new embeddings for spectral clustering.
        new_vectors_to_clusters = eigen_vectors.T @ sqrtM

        # Initialize kmeans klustering model.
        self.clustering_model = KMeans(
            n_clusters=self.nb_clusters,
            max_iter=10000,
            random_state=self.random_seed,
        )

        # Run kmeans clustering model.
        self.clustering_model.fit_predict(
            X=new_vectors_to_clusters
        )

        # Get prediction of kmeans clustering model.
        list_of_clusters: List[int] = self.clustering_model.labels_.tolist()

        # Define the dictionary of predicted clusters.
        predicted_clusters: Dict[str, int] = {
            data_ID: list_of_clusters[ID]
            for ID, data_ID in enumerate(self.list_of_data_IDs)
        }

        ###
        ### RENAME CLUSTERS BY ORDER
        ###

        # Define a map to be able to rename cluster IDs.
        mapping_of_old_ID_to_new_ID: Dict[int, int] = {}
        new_ID: int = 0
        for data_ID in self.list_of_data_IDs:
            if predicted_clusters[data_ID] not in mapping_of_old_ID_to_new_ID.keys():
                mapping_of_old_ID_to_new_ID[predicted_clusters[data_ID]] = new_ID
                new_ID += 1

        # Rename cluster IDs.
        predicted_clusters = {
            data_ID: mapping_of_old_ID_to_new_ID[predicted_clusters[data_ID]]
            for data_ID in self.list_of_data_IDs
        }

        # Return predicted clusters
        return predicted_clusters
    """



# -*- coding: utf-8 -*-

"""
* Name:         interactive-clustering/tests/constraints/test_abstract.py
* Description:  Unittests for the `constraints.abstract` module.
* Author:       Erwan Schild
* Created:      17/03/2021
* Licence:      CeCILL (https://cecill.info/licences.fr.html)
"""

# ==============================================================================
# IMPORT PYTHON DEPENDENCIES
# ==============================================================================

import pytest

from cognitivefactory.interactive_clustering.constraints.abstract import AbstractConstraintsManager


# ==============================================================================
# test_AbstractConstraintsManager_is_abstract
# ==============================================================================
def test_AbstractConstraintsManager_is_abstract():
    """
    Test that the `constraints.abstract.AbstractConstraintsManager` class is abstract.
    """

    # Check `TypeError` for initialization of `AbstractConstraintsManager`.
    with pytest.raises(TypeError, match="Can't instantiate abstract class"):
        AbstractConstraintsManager()



# -*- coding: utf-8 -*-

"""
* Name:         cognitivefactory.interactive_clustering.sampling.factory
* Description:  The factory method used to easily initialize a constraints sampling algorithm.
* Author:       Erwan Schild
* Created:      17/03/2021
* Licence:      CeCILL (https://cecill.info/licences.fr.html)
"""

# ==============================================================================
# IMPORT PYTHON DEPENDENCIES
# ==============================================================================

from cognitivefactory.interactive_clustering.sampling.abstract import (  # To use abstract interface.
    AbstractConstraintsSampling,
)
from cognitivefactory.interactive_clustering.sampling.clusters_based import (  # To use clusters based sampling.
    ClustersBasedConstraintsSampling,
)


# ==============================================================================
# SAMPLING FACTORY
# ==============================================================================
def sampling_factory(algorithm: str, **kargs) -> "AbstractConstraintsSampling":
    """
    A factory to create a new instance of a constraints sampling model.

    Args:
        algorithm (str): The identification of model to instantiate. Can be `"random"` or `"random_in_same_cluster"` or `"farthest_in_same_cluster"` or `"closest_in_different_clusters"`. Defaults to `"random"`
        **kargs (dict): Other parameters that can be used in the instantiation.

    Raises:
        ValueError: if `algorithm` is not implemented.

    Returns:
        AbstractConstraintsSampling: An instance of constraints sampling model.

    Examples:
        ```python
        # Import.
        from cognitivefactory.interactive_clustering.sampling.factory import sampling_factory

        # Create an instance of random sampler.
        sampler = sampling_factory(
            algorithm="random",
        )
        ```
    """

    # Check that the requested algorithm is implemented.
    if algorithm not in {
        "random",
        "random_in_same_cluster",
        "farthest_in_same_cluster",
        "closest_in_different_clusters",
    }:
        raise ValueError("The `algorithm` '" + str(algorithm) + "' is not implemented.")

    # Case of Random In Same Cluster Constraints Sampling.
    if algorithm == "random_in_same_cluster":
        return ClustersBasedConstraintsSampling(clusters_restriction="same_cluster", **kargs)

    # Case of Farthest In Same Cluster Constraints Sampling.
    if algorithm == "farthest_in_same_cluster":
        return ClustersBasedConstraintsSampling(
            distance_restriction="farthest_neighbors", clusters_restriction="same_cluster", **kargs
        )

    # Case of Closest In Different Clusters Constraints Sampling.
    if algorithm == "closest_in_different_clusters":
        return ClustersBasedConstraintsSampling(
            distance_restriction="closest_neighbors", clusters_restriction="different_clusters", **kargs
        )

    # Case of Random Constraints Sampling.
    ##if algorithm == "random":
    return ClustersBasedConstraintsSampling(**kargs)



# -*- coding: utf-8 -*-

"""
* Name:         cognitivefactory.interactive_clustering.clustering
* Description:  Constrained clustering module of the Interactive Clustering package.
* Author:       Erwan Schild
* Created:      17/03/2021
* Licence:      CeCILL-C License v1.0 (https://cecill.info/licences.fr.html)
"""



# -*- coding: utf-8 -*-

"""
* Name:         interactive-clustering/tests/utils/test_vectorization.py
* Description:  Unittests for the `utils.vectorization` module.
* Author:       Erwan Schild
* Created:      17/03/2021
* Licence:      CeCILL (https://cecill.info/licences.fr.html)
"""

# ==============================================================================
# IMPORT PYTHON DEPENDENCIES
# ==============================================================================

import pytest
from numpy import ndarray
from scipy.sparse import csr_matrix

from cognitivefactory.interactive_clustering.utils.vectorization import vectorize


# ==============================================================================
# test_vectorize_for_unimplemented_vectorizer
# ==============================================================================
def test_vectorize_for_unimplemented_vectorizer():
    """
    Test that the `utils.vectorization.vectorize` raises `ValueError` for unimplemented vectorizer.
    """

    # Check a unimplemented vectorizer.
    with pytest.raises(ValueError, match="`vectorizer_type`"):

        vectorize(
            dict_of_texts={
                "0": "comment signaler une perte de carte de paiement",
                "1": "quelle est la procedure pour chercher une carte de credit avalee",
                "2": "ma carte visa a un plafond de paiment trop bas puis je l augmenter",
            },
            vectorizer_type="unimplemented",
        )


# ==============================================================================
# test_vectorize_for_tfidf_vectorizer
# ==============================================================================
def test_vectorize_for_tfidf_vectorizer():
    """
    Test that the `utils.vectorization.vectorize` works for TFIDF vectorizer.
    """

    # Check a TFIDF vectorizer.
    dict_of_vectors = vectorize(
        dict_of_texts={
            "0": "comment signaler une perte de carte de paiement",
            "1": "quelle est la procedure pour chercher une carte de credit avalee",
            "2": "ma carte visa a un plafond de paiment trop bas puis je l augmenter",
        },
        vectorizer_type="tfidf",
    )

    # Assertions
    assert dict_of_vectors
    assert sorted(dict_of_vectors.keys()) == ["0", "1", "2"]
    assert isinstance(dict_of_vectors["0"], (ndarray, csr_matrix))
    assert isinstance(dict_of_vectors["1"], (ndarray, csr_matrix))
    assert isinstance(dict_of_vectors["2"], (ndarray, csr_matrix))


# ==============================================================================
# test_vectorize_for_uninstalled_spacy_language_model
# ==============================================================================
def test_vectorize_for_uninstalled_spacy_language_model():
    """
    Test that the `utils.vectorization.vectorize` raises `ValueError` for uninstalled spacy language model.
    """

    # Check a unimplemented vectorizer.
    with pytest.raises(ValueError, match="`spacy_language_model`"):

        vectorize(
            dict_of_texts={
                "0": "comment signaler une perte de carte de paiement",
                "1": "quelle est la procedure pour chercher une carte de credit avalee",
                "2": "ma carte visa a un plafond de paiment trop bas puis je l augmenter",
            },
            vectorizer_type="spacy",
            spacy_language_model="uninstalled",
        )


# ==============================================================================
# test_vectorize_for_installed_spacy_language_model
# ==============================================================================
def test_vectorize_for_installed_spacy_language_model():
    """
    Test that the `utils.vectorization.vectorize` works for an installed spacy language model.
    """

    # Check a SPACY vectorizer.
    dict_of_vectors = vectorize(
        dict_of_texts={
            "0": "hello how are you",
            "1": "hello how old are you",
            "2": "hello where do you live",
        },
        vectorizer_type="spacy",
        spacy_language_model="en_core_web_sm",
    )

    # Assertions
    assert dict_of_vectors


# ==============================================================================
# test_vectorize_for_spacy_vectorizer
# ==============================================================================
def test_vectorize_for_spacy_vectorizer():
    """
    Test that the `utils.vectorization.vectorize` works for SPACY vectorizer.
    """

    # Check a SPACY vectorizer.
    dict_of_vectors = vectorize(
        dict_of_texts={
            "0": "comment signaler une perte de carte de paiement",
            "1": "quelle est la procedure pour chercher une carte de credit avalee",
            "2": "ma carte visa a un plafond de paiment trop bas puis je l augmenter",
        },
        vectorizer_type="spacy",
    )

    # Assertions
    assert dict_of_vectors
    assert sorted(dict_of_vectors.keys()) == ["0", "1", "2"]
    assert isinstance(dict_of_vectors["0"], (ndarray, csr_matrix))
    assert isinstance(dict_of_vectors["1"], (ndarray, csr_matrix))
    assert isinstance(dict_of_vectors["2"], (ndarray, csr_matrix))



# -*- coding: utf-8 -*-

"""
* Name:         cognitivefactory.interactive_clustering.clustering.factory
* Description:  The factory method used to easily initialize a constrained clustering algorithm.
* Author:       Erwan Schild
* Created:      17/03/2021
* Licence:      CeCILL-C License v1.0 (https://cecill.info/licences.fr.html)
"""

# ==============================================================================
# IMPORT PYTHON DEPENDENCIES
# ==============================================================================

from cognitivefactory.interactive_clustering.clustering.abstract import (  # To use abstract interface.
    AbstractConstrainedClustering,
)
from cognitivefactory.interactive_clustering.clustering.hierarchical import (  # To use hierarchical constrained clustering implementation.
    HierarchicalConstrainedClustering,
)
from cognitivefactory.interactive_clustering.clustering.kmeans import (  # To use kmeans constrained clustering implementation.
    KMeansConstrainedClustering,
)
from cognitivefactory.interactive_clustering.clustering.spectral import (  # To use spectral constrained clustering implementation.
    SpectralConstrainedClustering,
)


# ==============================================================================
# CLUSTERING FACTORY
# ==============================================================================
def clustering_factory(algorithm: str = "kmeans", **kargs) -> "AbstractConstrainedClustering":
    """
    A factory to create a new instance of a constrained clustering model.

    Args:
        algorithm (str): The identification of model to instantiate. Can be `"hierarchical"` or `"kmeans"` or `"spectral"`. Defaults to `"kmeans"`.
        **kargs (dict): Other parameters that can be used in the instantiation.

    Raises:
        ValueError: if `algorithm` is not implemented.

    Returns:
        AbstractConstraintsClustering: An instance of constrained clustering model.

    Examples:
        ```python
        # Import.
        from cognitivefactory.interactive_clustering.clustering.factory import clustering_factory

        # Create an instance of kmeans.
        clustering_model = clustering_factory(
            algorithm="kmeans",
            model="COP",
            random_seed=42,
        )
        ```
    """

    # Check that the requested algorithm is implemented.
    if algorithm not in {"hierarchical", "kmeans", "spectral"}:
        raise ValueError("The `algorithm` '" + str(algorithm) + "' is not implemented.")

    # Case of Hierachical Constrained Clustering.
    if algorithm == "hierarchical":

        return HierarchicalConstrainedClustering(**kargs)

    # Case of Spectral Constrained Clustering.
    if algorithm == "spectral":

        return SpectralConstrainedClustering(**kargs)

    # Case of KMeans Constrained Clustering.
    ## if algorithm=="kmeans":
    return KMeansConstrainedClustering(**kargs)



# -*- coding: utf-8 -*-

"""
* Name:         cognitivefactory.interactive_clustering.clustering.abstract
* Description:  The abstract class used to define constrained clustering algorithms.
* Author:       Erwan Schild
* Created:      17/03/2021
* Licence:      CeCILL-C License v1.0 (https://cecill.info/licences.fr.html)
"""

# ==============================================================================
# IMPORT PYTHON DEPENDENCIES
# ==============================================================================

from abc import ABC, abstractmethod  # To define an abstract class.
from typing import Dict, Union  # To type Python code (mypy).

from numpy import ndarray  # To handle matrix and vectors.
from scipy.sparse import csr_matrix  # To handle matrix and vectors.

from cognitivefactory.interactive_clustering.constraints.abstract import (  # To manage constraints.
    AbstractConstraintsManager,
)


# ==============================================================================
# ABSTRACT CONSTRAINED CLUSTERING
# ==============================================================================
class AbstractConstrainedClustering(ABC):
    """
    Abstract class that is used to define constrained clustering algorithms.
    The main inherited method is `cluster`.

    References:
        - Survey on Constrained Clustering : `Lampert, T., T.-B.-H. Dao, B. Lafabregue, N. Serrette, G. Forestier, B. Cremilleux, C. Vrain, et P. Gancarski (2018). Constrained distance based clustering for time-series : a comparative and experimental study. Data Mining and Knowledge Discovery 32(6), 1663–1707.`
    """

    # ==============================================================================
    # ABSTRACT METHOD - CLUSTER
    # ==============================================================================
    @abstractmethod
    def cluster(
        self,
        constraints_manager: AbstractConstraintsManager,
        vectors: Dict[str, Union[ndarray, csr_matrix]],
        nb_clusters: int,
        verbose: bool = False,
        **kargs,
    ) -> Dict[str, int]:
        """
        (ABSTRACT METHOD)
        An abstract method that represents the main method used to cluster data.

        Args:
            constraints_manager (AbstractConstraintsManager): A constraints manager over data IDs that will force clustering to respect some conditions during computation.
            vectors (Dict[str,Union[ndarray,csr_matrix]]): The representation of data vectors. The keys of the dictionary represents the data IDs. This keys have to refer to the list of data IDs managed by the `constraints_manager`. The value of the dictionary represent the vector of each data. Vectors can be dense (`numpy.ndarray`) or sparse (`scipy.sparse.csr_matrix`).
            nb_clusters (int): The number of clusters to compute. #TODO Set defaults to None with elbow method or other method ?
            verbose (bool, optional): Enable verbose output. Defaults to `False`.
            **kargs (dict): Other parameters that can be used in the clustering.

        Raises:
            ValueError: if `vectors` and `constraints_manager` are incompatible, or if some parameters are incorrectly set.

        Returns:
            Dict[str,int]: A dictionary that contains the predicted cluster for each data ID.
        """


# ==============================================================================
# RENAME CLUSTERS BY ORDER
# ==============================================================================
def rename_clusters_by_order(
    clusters: Dict[str, int],
) -> Dict[str, int]:
    """
    Rename cluster ID to be ordered by data IDs.

    Args:
        clusters (Dict[str, int]): The dictionary of clusters.

    Returns:
        Dict[str, int]: The sorted dictionary of clusters.
    """

    # Get `list_of_data_IDs`.
    list_of_data_IDs = sorted(clusters.keys())

    # Define a map to be able to rename cluster IDs.
    mapping_of_old_ID_to_new_ID: Dict[int, int] = {}
    new_ID: int = 0
    for data_ID in list_of_data_IDs:  # , cluster_ID in clusters.items():
        if clusters[data_ID] not in mapping_of_old_ID_to_new_ID.keys():
            mapping_of_old_ID_to_new_ID[clusters[data_ID]] = new_ID
            new_ID += 1

    # Rename cluster IDs.
    new_clusters = {
        data_ID_to_assign: mapping_of_old_ID_to_new_ID[clusters[data_ID_to_assign]]
        for data_ID_to_assign in list_of_data_IDs
    }

    # Return the new ordered clusters
    return new_clusters



# -*- coding: utf-8 -*-

"""
* Name:         interactive-clustering/tests/constraints/test_binary.py
* Description:  Unittests for the `constraints.binary` module.
* Author:       Erwan Schild
* Created:      17/03/2021
* Licence:      CeCILL (https://cecill.info/licences.fr.html)
"""

# ==============================================================================
# IMPORT PYTHON DEPENDENCIES
# ==============================================================================

import pytest

from cognitivefactory.interactive_clustering.constraints.binary import BinaryConstraintsManager


# ==============================================================================
# test_BinaryConstraintsManager_init
# ==============================================================================
def test_BinaryConstraintsManager_init():
    """
    Test that the `__init__` method of the `constraints.binary.BinaryConstraintsManager` class works.
    """

    # Initialize an empty binaray constraints manager.
    constraints_manager_1 = BinaryConstraintsManager(
        list_of_data_IDs=[],
    )
    assert not constraints_manager_1.get_list_of_managed_data_IDs()

    # Initialize a classic binaray constraints manager.
    constraints_manager_2 = BinaryConstraintsManager(
        list_of_data_IDs=["first", "second", "third"],
    )
    assert constraints_manager_2.get_list_of_managed_data_IDs() == ["first", "second", "third"]

    # Initialize a binaray constraints manager with duplicates in `list_of_data_IDs`.
    with pytest.raises(ValueError, match="`list_of_data_IDs`"):
        BinaryConstraintsManager(
            list_of_data_IDs=["first", "second", "same", "same"],
        )


# ==============================================================================
# test_BinaryConstraintsManager_add_data_ID_with_incorrect_data_ID
# ==============================================================================
def test_BinaryConstraintsManager_add_data_ID_with_incorrect_data_ID():
    """
    Test that the `add_data_ID` method of the `constraints.binary.BinaryConstraintsManager` class raises `ValueError` for an incorrect data ID.
    """

    # Initialize a classic binaray constraints manager.
    constraints_manager = BinaryConstraintsManager(
        list_of_data_IDs=["first", "second", "third"],
    )

    # Try to add `"second"` data ID.
    with pytest.raises(ValueError, match="`data_ID`"):
        constraints_manager.add_data_ID(
            data_ID="second",
        )

    # Run assertions.
    assert constraints_manager.get_list_of_managed_data_IDs() == ["first", "second", "third"]


# ==============================================================================
# test_BinaryConstraintsManager_add_data_ID_with_correct_parameters
# ==============================================================================
def test_BinaryConstraintsManager_add_data_ID_with_correct_parameters():
    """
    Test that the `add_data_ID` method of the `constraints.binary.BinaryConstraintsManager` class returns `True` for a correct parameters.
    """

    # Initialize a classic binaray constraints manager.
    constraints_manager = BinaryConstraintsManager(
        list_of_data_IDs=["first", "second", "third"],
    )

    # Try to add `"fourth"` data ID.
    assert (
        constraints_manager.add_data_ID(
            data_ID="fourth",
        )
        is True
    )

    # Run assertions.
    assert constraints_manager.get_list_of_managed_data_IDs() == ["first", "second", "third", "fourth"]
    assert (
        constraints_manager.get_inferred_constraint(
            data_ID1="first",
            data_ID2="fourth",
        )
        is None
    )
    assert (
        constraints_manager.get_inferred_constraint(
            data_ID1="second",
            data_ID2="fourth",
        )
        is None
    )
    assert (
        constraints_manager.get_inferred_constraint(
            data_ID1="third",
            data_ID2="fourth",
        )
        is None
    )
    assert (
        constraints_manager.get_inferred_constraint(
            data_ID1="fourth",
            data_ID2="fourth",
        )
        == "MUST_LINK"
    )


# ==============================================================================
# test_BinaryConstraintsManager_delete_data_ID_with_incorrect_data_ID
# ==============================================================================
def test_BinaryConstraintsManager_delete_data_ID_with_incorrect_data_ID():
    """
    Test that the `delete_data_ID` method of the `constraints.binary.BinaryConstraintsManager` class raises `ValueError` for an incorrect data ID.
    """

    # Initialize a classic binaray constraints manager.
    constraints_manager = BinaryConstraintsManager(
        list_of_data_IDs=["first", "second", "third"],
    )

    # Try to delete `"unknown"` data ID.
    with pytest.raises(ValueError, match="`data_ID`"):
        constraints_manager.delete_data_ID(
            data_ID="unknown",
        )

    # Run assertions.
    assert constraints_manager.get_list_of_managed_data_IDs() == ["first", "second", "third"]


# ==============================================================================
# test_BinaryConstraintsManager_delete_data_ID_with_correct_parameters
# ==============================================================================
def test_BinaryConstraintsManager_delete_data_ID_with_correct_parameters():
    """
    Test that the `delete_data_ID` method of the `constraints.binary.BinaryConstraintsManager` class returns `True` for a correct parameters.
    """

    # Initialize a classic binaray constraints manager.
    constraints_manager = BinaryConstraintsManager(
        list_of_data_IDs=["first", "second", "third"],
    )

    # Try to delete `"second"` data ID.
    assert (
        constraints_manager.delete_data_ID(
            data_ID="second",
        )
        is True
    )

    # Run assertions.
    assert constraints_manager.get_list_of_managed_data_IDs() == ["first", "third"]


# ==============================================================================
# test_BinaryConstraintsManager_get_list_of_managed_data_IDs
# ==============================================================================

# Not implemented because `get_list_of_managed_data_IDs` is already tested in other unittests.


# ==============================================================================
# test_BinaryConstraintsManager_add_constraint_with_incorrect_data_ID
# ==============================================================================
def test_BinaryConstraintsManager_add_constraint_with_incorrect_data_ID():
    """
    Test that the `add_constraint` method of the `constraints.binary.BinaryConstraintsManager` class raises `ValueError` for incorrect data IDs.
    """

    # Initialize a classic binaray constraints manager.
    constraints_manager = BinaryConstraintsManager(
        list_of_data_IDs=["first", "second", "third"],
    )

    # Try to add the constraint with `"unknown"` data ID in `"data_ID1"`.
    with pytest.raises(ValueError, match="`data_ID1`"):
        constraints_manager.add_constraint(
            data_ID1="unknown",
            data_ID2="second",
            constraint_type="MUST_LINK",
        )

    # Try to add the constraint with `"unknown"` data ID in `"data_ID2"`.
    with pytest.raises(ValueError, match="`data_ID2`"):
        constraints_manager.add_constraint(
            data_ID1="first",
            data_ID2="unknown",
            constraint_type="CANNOT_LINK",
        )


# ==============================================================================
# test_BinaryConstraintsManager_add_constraint_with_incorrect_constraint_type
# ==============================================================================
def test_BinaryConstraintsManager_add_constraint_with_incorrect_constraint_type():
    """
    Test that the `add_constraint` method of the `constraints.binary.BinaryConstraintsManager` class raises `ValueError` for an incorrect constraint type.
    """

    # Initialize a classic binaray constraints manager.
    constraints_manager = BinaryConstraintsManager(
        list_of_data_IDs=["first", "second", "third"],
    )

    # Try to add the constraint with `"UNKNOWN_LINK"` constraint type in `"constraint_type"`.
    with pytest.raises(ValueError, match="`constraint_type`"):
        constraints_manager.add_constraint(
            data_ID1="first",
            data_ID2="second",
            constraint_type="UNKNOWN_LINK",
        )


# ==============================================================================
# test_BinaryConstraintsManager_add_constraint_with_already_linked_data_IDs
# ==============================================================================
def test_BinaryConstraintsManager_add_constraint_with_already_linked_data_IDs():
    """
    Test that the `add_constraint` method of the `constraints.binary.BinaryConstraintsManager` class works for already linked data IDs.
    """

    # Initialize a classic binaray constraints manager.
    constraints_manager = BinaryConstraintsManager(
        list_of_data_IDs=["first", "second", "third"],
    )
    constraints_manager.add_constraint(
        data_ID1="first",
        data_ID2="second",
        constraint_type="MUST_LINK",
    )
    constraints_manager.add_constraint(
        data_ID1="first",
        data_ID2="third",
        constraint_type="CANNOT_LINK",
    )

    # Try to add `"MUST_LINK"` data ID between `"first"` and `"second"` data IDs.
    assert (
        constraints_manager.add_constraint(
            data_ID1="first",
            data_ID2="second",
            constraint_type="MUST_LINK",
        )
        is True
    )

    # Try to add `"CANNOT_LINK"` data ID between `"first"` and `"second"` data IDs.
    with pytest.raises(ValueError, match="`constraint_type`"):
        constraints_manager.add_constraint(
            data_ID1="first",
            data_ID2="second",
            constraint_type="CANNOT_LINK",
        )


# ==============================================================================
# test_BinaryConstraintsManager_add_constraint_with_not_already_linked_data_IDs
# ==============================================================================
def test_BinaryConstraintsManager_add_constraint_with_not_already_linked_data_IDs():
    """
    Test that the `add_constraint` method of the `constraints.binary.BinaryConstraintsManager` class works for not already linked data IDs.
    """

    # Initialize a classic binaray constraints manager.
    constraints_manager = BinaryConstraintsManager(
        list_of_data_IDs=["first", "second", "third"],
    )

    # Try to add `"MUST_LINK"` data ID between `"first"` and `"second"` data IDs.
    assert (
        constraints_manager.add_constraint(
            data_ID1="first",
            data_ID2="second",
            constraint_type="MUST_LINK",
        )
        is True
    )
    assert (
        constraints_manager.get_added_constraint(
            data_ID1="first",
            data_ID2="second",
        )
        == ("MUST_LINK", 1.0)
    )

    # Try to add `"CANNOT_LINK"` data ID between `"first"` and `"third"` data IDs.
    assert (
        constraints_manager.add_constraint(
            data_ID1="first",
            data_ID2="third",
            constraint_type="CANNOT_LINK",
        )
        is True
    )
    assert (
        constraints_manager.get_added_constraint(
            data_ID1="first",
            data_ID2="third",
        )
        == ("CANNOT_LINK", 1.0)
    )

    # Try to add `"MUST_LINK"` data ID between `"second"` and `"third"` data IDs.
    with pytest.raises(ValueError, match="`constraint_type`"):
        constraints_manager.add_constraint(
            data_ID1="second",
            data_ID2="third",
            constraint_type="MUST_LINK",
        )

    # Try to add `"CANNOT_LINK"` data ID between `"second"` and `"third"` data IDs.
    assert (
        constraints_manager.add_constraint(
            data_ID1="second",
            data_ID2="third",
            constraint_type="CANNOT_LINK",
        )
        is True
    )
    assert (
        constraints_manager.get_added_constraint(
            data_ID1="second",
            data_ID2="third",
        )
        == ("CANNOT_LINK", 1.0)
    )


# ==============================================================================
# test_BinaryConstraintsManager_delete_constraint_with_incorrect_data_IDs
# ==============================================================================
def test_BinaryConstraintsManager_delete_constraint_with_incorrect_data_IDs():
    """
    Test that the `delete_constraint` method of the `constraints.binary.BinaryConstraintsManager` class raises `ValueError` for incorrect data IDs.
    """

    # Initialize a classic binaray constraints manager.
    constraints_manager = BinaryConstraintsManager(
        list_of_data_IDs=["first", "second", "third"],
    )

    # Try to delete constraint with `"unknown"` data ID in `"data_ID1"`.
    with pytest.raises(ValueError, match="`data_ID1`"):
        constraints_manager.delete_constraint(
            data_ID1="unknown",
            data_ID2="second",
        )

    # Try to delete constraint with `"unknown"` data ID in `"data_ID2"`.
    with pytest.raises(ValueError, match="`data_ID2`"):
        constraints_manager.delete_constraint(
            data_ID1="first",
            data_ID2="unknown",
        )


# ==============================================================================
# test_BinaryConstraintsManager_delete_constraint_with_correct_parameters
# ==============================================================================
def test_BinaryConstraintsManager_delete_constraint_with_correct_parameters():
    """
    Test that the `delete_constraint` method of the `constraints.binary.BinaryConstraintsManager` returns `True` for correct parameters.
    """

    # Initialize a classic binaray constraints manager.
    constraints_manager = BinaryConstraintsManager(
        list_of_data_IDs=["first", "second", "third"],
    )
    constraints_manager.add_constraint(
        data_ID1="first",
        data_ID2="second",
        constraint_type="MUST_LINK",
    )
    constraints_manager.add_constraint(
        data_ID1="first",
        data_ID2="third",
        constraint_type="CANNOT_LINK",
    )

    # Try to delete constraint betwenn `"second"` and `"third"` data IDs.
    assert (
        constraints_manager.delete_constraint(
            data_ID1="second",
            data_ID2="third",
        )
        is True
    )
    assert (
        constraints_manager.get_added_constraint(
            data_ID1="second",
            data_ID2="third",
        )
        is None
    )
    assert (
        constraints_manager.get_inferred_constraint(
            data_ID1="second",
            data_ID2="third",
        )
        == "CANNOT_LINK"
    )  # transitivity activated !

    # Try to delete constraint betwenn `"first"` and `"third"` data IDs.
    assert (
        constraints_manager.delete_constraint(
            data_ID1="first",
            data_ID2="third",
        )
        is True
    )
    assert (
        constraints_manager.get_added_constraint(
            data_ID1="first",
            data_ID2="third",
        )
        is None
    )
    assert (
        constraints_manager.get_added_constraint(
            data_ID1="second",
            data_ID2="third",
        )
        is None
    )  # No more transitivity


# ==============================================================================
# test_BinaryConstraintsManager_get_added_constraint_with_incorrect_data_ID
# ==============================================================================
def test_BinaryConstraintsManager_get_added_constraint_with_incorrect_data_ID():
    """
    Test that the `get_added_constraint` method of the `constraints.binary.BinaryConstraintsManager` class raises `ValueError` for an incorrect data ID.
    """

    # Initialize a classic binaray constraints manager.
    constraints_manager = BinaryConstraintsManager(
        list_of_data_IDs=["first", "second", "third"],
    )
    constraints_manager.add_constraint(
        data_ID1="first",
        data_ID2="second",
        constraint_type="MUST_LINK",
    )
    constraints_manager.add_constraint(
        data_ID1="first",
        data_ID2="third",
        constraint_type="CANNOT_LINK",
    )

    # Try to get the constraint with `"unknown"` data ID in `"data_ID1"`.
    with pytest.raises(ValueError, match="`data_ID1`"):
        constraints_manager.get_added_constraint(
            data_ID1="unknown",
            data_ID2="second",
        )

    # Try to get the constraint with `"unknown"` data ID in `"data_ID2"`.
    with pytest.raises(ValueError, match="`data_ID2`"):
        constraints_manager.get_added_constraint(
            data_ID1="first",
            data_ID2="unknown",
        )


# ==============================================================================
# test_BinaryConstraintsManager_get_added_constraint_with_correct_parameters
# ==============================================================================
def test_BinaryConstraintsManager_get_added_constraint_with_correct_parameters():
    """
    Test that the `get_added_constraint` method of the `constraints.binary.BinaryConstraintsManager` class returns `True` for correct parameters.
    """

    # Initialize a classic binaray constraints manager.
    constraints_manager = BinaryConstraintsManager(
        list_of_data_IDs=["first", "second", "third"],
    )
    constraints_manager.add_constraint(
        data_ID1="first",
        data_ID2="second",
        constraint_type="MUST_LINK",
    )
    constraints_manager.add_constraint(
        data_ID1="first",
        data_ID2="third",
        constraint_type="CANNOT_LINK",
    )

    # Try to get the constraint betwwen `"first"` and `"second"`.
    assert (
        constraints_manager.get_added_constraint(
            data_ID1="first",
            data_ID2="second",
        )
        == ("MUST_LINK", 1.0)
    )

    # Try to get the constraint betwwen `"first"` and `"third"`.
    assert (
        constraints_manager.get_added_constraint(
            data_ID1="first",
            data_ID2="third",
        )
        == ("CANNOT_LINK", 1.0)
    )


# ==============================================================================
# test_BinaryConstraintsManager_get_inferred_constraint_with_incorrect_data_ID
# ==============================================================================
def test_BinaryConstraintsManager_get_inferred_constraint_with_incorrect_data_ID():
    """
    Test that the `get_inferred_constraint` method of the `constraints.binary.BinaryConstraintsManager` class raises `ValueError` for an incorrect data ID.
    """

    # Initialize a classic binaray constraints manager.
    constraints_manager = BinaryConstraintsManager(
        list_of_data_IDs=["first", "second", "third"],
    )

    # Try `get_inferred_constraint` method with `"unknown"` data ID in `data_ID1`.
    with pytest.raises(ValueError, match="`data_ID1`"):
        constraints_manager.get_inferred_constraint(
            data_ID1="unknown",
            data_ID2="second",
        )

    # Try `get_inferred_constraint` method with `"unknown"` data ID in `data_ID2`.
    with pytest.raises(ValueError, match="`data_ID2`"):
        constraints_manager.get_inferred_constraint(
            data_ID1="first",
            data_ID2="unknown",
        )


# ==============================================================================
# test_BinaryConstraintsManager_get_inferred_constraint_with_correct_parameter
# ==============================================================================
def test_BinaryConstraintsManager_get_inferred_constraint_with_correct_parameter():
    """
    Test that the `get_inferred_constraint` method of the `constraints.binary.BinaryConstraintsManager` class works for a correct data ID.
    """

    # Initialize a classic binaray constraints manager.
    constraints_manager = BinaryConstraintsManager(
        list_of_data_IDs=["first", "second", "third"],
    )

    # Add `"MUST_LINK"` constraint between `"first"` and `"second"`.
    constraints_manager.add_constraint(
        data_ID1="first",
        data_ID2="second",
        constraint_type="MUST_LINK",
    )

    # Try to check the link bewteen `"first"` and `"second"`data IDs.
    assert (
        constraints_manager.get_inferred_constraint(
            data_ID1="first",
            data_ID2="second",
        )
        == "MUST_LINK"
    )

    # Try to check the link bewteen `"first"` and `"third"`data IDs.
    assert (
        constraints_manager.get_inferred_constraint(
            data_ID1="first",
            data_ID2="third",
        )
        is None
    )


# ==============================================================================
# test_BinaryConstraintsManager_get_connected_components
# ==============================================================================
def test_BinaryConstraintsManager_get_connected_components():
    """
    Test that the `get_connected_components` method of the `constraints.binary.BinaryConstraintsManager` class works.
    """

    # Initialize a classic binaray constraints manager.
    constraints_manager = BinaryConstraintsManager(
        list_of_data_IDs=["first", "second", "third", "fourth"],
    )

    # Add `"MUST_LINK"` constraint between `"first"` and `"second"`.
    constraints_manager.add_constraint(
        data_ID1="first",
        data_ID2="second",
        constraint_type="MUST_LINK",
    )
    # Add `"MUST_LINK"` constraint between `"second"` and `"third"`.
    constraints_manager.add_constraint(
        data_ID1="second",
        data_ID2="third",
        constraint_type="MUST_LINK",
    )
    # Add `"CANNOT_LINK"` constraint between `"second"` and `"fourth"`.
    constraints_manager.add_constraint(
        data_ID1="second",
        data_ID2="fourth",
        constraint_type="CANNOT_LINK",
    )

    # Try to get the list of connected components.
    connected_component = constraints_manager.get_connected_components()
    assert connected_component == [["first", "second", "third"], ["fourth"]]


# ==============================================================================
# test_BinaryConstraintsManager_check_completude_of_constraints
# ==============================================================================
def test_BinaryConstraintsManager_check_completude_of_constraints():
    """
    Test that the `check_completude_of_constraints` method of the `constraints.binary.BinaryConstraintsManager` class works.
    """

    # Initialize a classic binaray constraints manager.
    constraints_manager = BinaryConstraintsManager(
        list_of_data_IDs=["first", "second", "third", "fourth"],
    )
    assert constraints_manager.check_completude_of_constraints() is False

    # Add `"MUST_LINK"` constraint between `"first"` and `"second"`.
    constraints_manager.add_constraint(
        data_ID1="first",
        data_ID2="second",
        constraint_type="MUST_LINK",
    )
    assert constraints_manager.check_completude_of_constraints() is False

    # Add `"MUST_LINK"` constraint between `"second"` and `"third"`.
    constraints_manager.add_constraint(
        data_ID1="second",
        data_ID2="third",
        constraint_type="MUST_LINK",
    )
    assert constraints_manager.check_completude_of_constraints() is False

    # Add `"CANNOT_LINK"` constraint between `"second"` and `"fourth"`.
    constraints_manager.add_constraint(
        data_ID1="second",
        data_ID2="fourth",
        constraint_type="CANNOT_LINK",
    )
    assert constraints_manager.check_completude_of_constraints() is True


# ==============================================================================
# test_BinaryConstraintsManager_get_min_and_max_number_of_clusters
# ==============================================================================
def test_BinaryConstraintsManager_get_min_and_max_number_of_clusters():
    """
    Test that the `get_min_and_max_number_of_clusters` method of the `constraints.binary.BinaryConstraintsManager` class works.
    """

    # Initialize a classic binaray constraints manager.
    constraints_manager = BinaryConstraintsManager(
        list_of_data_IDs=["first", "second", "third", "fourth", "fifth", "sixth"],
    )
    min, max = constraints_manager.get_min_and_max_number_of_clusters()
    assert min == 2
    assert max == 6

    # Add `"MUST_LINK"` constraint between `"first"` and `"second"`.
    constraints_manager.add_constraint(
        data_ID1="first",
        data_ID2="second",
        constraint_type="MUST_LINK",
    )
    min, max = constraints_manager.get_min_and_max_number_of_clusters()
    assert min == 2
    assert max == 5

    # Add `"CANNOT_LINK"` constraint between `"second"` and `"third"`.
    constraints_manager.add_constraint(
        data_ID1="second",
        data_ID2="third",
        constraint_type="CANNOT_LINK",
    )
    min, max = constraints_manager.get_min_and_max_number_of_clusters()
    assert min == 2
    assert max == 5

    # Add `"CANNOT_LINK"` constraint between `"second"` and `"fourth"`.
    constraints_manager.add_constraint(
        data_ID1="second",
        data_ID2="fourth",
        constraint_type="CANNOT_LINK",
    )
    min, max = constraints_manager.get_min_and_max_number_of_clusters()
    assert min == 2
    assert max == 5

    # Add `"CANNOT_LINK"` constraint between `"third"` and `"fourth"`.
    constraints_manager.add_constraint(
        data_ID1="third",
        data_ID2="fourth",
        constraint_type="CANNOT_LINK",
    )
    min, max = constraints_manager.get_min_and_max_number_of_clusters()
    assert min == 3  # 2.99999...
    assert max == 5

    # Add `"MUST_LINK"` constraint between `"fourth"` and `"fifth"`.
    constraints_manager.add_constraint(
        data_ID1="fourth",
        data_ID2="fifth",
        constraint_type="MUST_LINK",
    )
    min, max = constraints_manager.get_min_and_max_number_of_clusters()
    assert min == 3  # 2.99999...
    assert max == 4

    # Add `"MUST_LINK"` constraint between `"fourth"` and `"sixth"`.
    constraints_manager.add_constraint(
        data_ID1="fourth",
        data_ID2="sixth",
        constraint_type="MUST_LINK",
    )
    min, max = constraints_manager.get_min_and_max_number_of_clusters()
    assert min == 3  # 2.99999...
    assert max == 3
    assert constraints_manager.check_completude_of_constraints() is True


# ==============================================================================
# test_BinaryConstraintsManager_check_symetry_after_constraint_addition
# ==============================================================================
def test_BinaryConstraintsManager_check_symetry_after_constraint_addition():
    """
    Test that the `add_constraint` method of the `constraints.binary.BinaryConstraintsManager` class is symetric.
    """

    # Initialize a classic binaray constraints manager.
    constraints_manager = BinaryConstraintsManager(
        list_of_data_IDs=["first", "second", "third"],
    )

    # Add `"MUST_LINK"` constraint between `"first"` and `"second"`.
    constraints_manager.add_constraint(
        data_ID1="first",
        data_ID2="second",
        constraint_type="MUST_LINK",
    )
    # Add `"CANNOT_LINK"` constraint between `"first"` and `"third"`.
    constraints_manager.add_constraint(
        data_ID1="first",
        data_ID2="third",
        constraint_type="CANNOT_LINK",
    )

    # Run assertion on : ML(1,2) => ML(2,1).
    assert (
        constraints_manager.get_added_constraint(
            data_ID1="first",
            data_ID2="second",
        )
        == ("MUST_LINK", 1.0)
    )
    assert (
        constraints_manager.get_added_constraint(
            data_ID1="second",
            data_ID2="first",
        )
        == ("MUST_LINK", 1.0)
    )
    assert (
        constraints_manager.get_inferred_constraint(
            data_ID1="first",
            data_ID2="second",
        )
        == "MUST_LINK"
    )
    assert (
        constraints_manager.get_inferred_constraint(
            data_ID1="second",
            data_ID2="first",
        )
        == "MUST_LINK"
    )

    # Run assertion on : CL(1,3) => CL(3,1).
    assert (
        constraints_manager.get_added_constraint(
            data_ID1="first",
            data_ID2="third",
        )
        == ("CANNOT_LINK", 1.0)
    )
    assert (
        constraints_manager.get_added_constraint(
            data_ID1="third",
            data_ID2="first",
        )
        == ("CANNOT_LINK", 1.0)
    )
    assert (
        constraints_manager.get_inferred_constraint(
            data_ID1="first",
            data_ID2="third",
        )
        == "CANNOT_LINK"
    )
    assert (
        constraints_manager.get_inferred_constraint(
            data_ID1="third",
            data_ID2="first",
        )
        == "CANNOT_LINK"
    )


# ==============================================================================
# test_BinaryConstraintsManager_check_transitivity_after_constraint_addition
# ==============================================================================
def test_BinaryConstraintsManager_check_transitivity_after_constraint_addition():
    """
    Test that the `add_constraint` method of the `constraints.binary.BinaryConstraintsManager` class is transitive.
    """

    # Initialize a classic binaray constraints manager.
    constraints_manager = BinaryConstraintsManager(
        list_of_data_IDs=["first", "second", "third", "fourth"],
    )

    # Add `"MUST_LINK"` constraint between `"first"` and `"second"`.
    constraints_manager.add_constraint(
        data_ID1="first",
        data_ID2="second",
        constraint_type="MUST_LINK",
    )
    # Add `"MUST_LINK"` constraint between `"second"` and `"third"`.
    constraints_manager.add_constraint(
        data_ID1="second",
        data_ID2="third",
        constraint_type="MUST_LINK",
    )
    # Add `"CANNOT_LINK"` constraint between `"second"` and `"fourth"`.
    constraints_manager.add_constraint(
        data_ID1="second",
        data_ID2="fourth",
        constraint_type="CANNOT_LINK",
    )

    # Run assertion on : ML(1,2)+ML(2,3) => ML(1,3) AND ML(3,1).
    assert (
        constraints_manager.get_added_constraint(
            data_ID1="first",
            data_ID2="third",
        )
        is None
    )
    assert (
        constraints_manager.get_added_constraint(
            data_ID1="third",
            data_ID2="first",
        )
        is None
    )
    assert (
        constraints_manager.get_inferred_constraint(
            data_ID1="first",
            data_ID2="third",
        )
        == "MUST_LINK"
    )
    assert (
        constraints_manager.get_inferred_constraint(
            data_ID1="third",
            data_ID2="first",
        )
        == "MUST_LINK"
    )

    # Run assertion on : ML(1,2)+CL(2,4) => CL(1,4) AND CL(4,1).
    assert (
        constraints_manager.get_added_constraint(
            data_ID1="first",
            data_ID2="fourth",
        )
        is None
    )
    assert (
        constraints_manager.get_added_constraint(
            data_ID1="fourth",
            data_ID2="first",
        )
        is None
    )
    assert (
        constraints_manager.get_inferred_constraint(
            data_ID1="first",
            data_ID2="fourth",
        )
        == "CANNOT_LINK"
    )
    assert (
        constraints_manager.get_inferred_constraint(
            data_ID1="fourth",
            data_ID2="first",
        )
        == "CANNOT_LINK"
    )


# ==============================================================================
# test_BinaryConstraintsManager_check_symetry_after_constraint_deletion
# ==============================================================================
def test_BinaryConstraintsManager_check_symetry_after_constraint_deletion():
    """
    Test that the `delete_constraint` method of the `constraints.binary.BinaryConstraintsManager` class is symetric.
    """

    # Initialize a classic binaray constraints manager.
    constraints_manager = BinaryConstraintsManager(
        list_of_data_IDs=["first", "second", "third"],
    )

    # Add `"MUST_LINK"` constraint between `"first"` and `"second"`.
    constraints_manager.add_constraint(
        data_ID1="first",
        data_ID2="second",
        constraint_type="MUST_LINK",
    )
    # Delete constraint between `"second"` and `"first"`.
    constraints_manager.delete_constraint(
        data_ID1="second",
        data_ID2="first",
    )

    # Run assertion (no more symetry).
    assert (
        constraints_manager.get_added_constraint(
            data_ID1="first",
            data_ID2="second",
        )
        is None
    )
    assert (
        constraints_manager.get_added_constraint(
            data_ID1="second",
            data_ID2="first",
        )
        is None
    )
    assert (
        constraints_manager.get_inferred_constraint(
            data_ID1="first",
            data_ID2="second",
        )
        is None
    )
    assert (
        constraints_manager.get_inferred_constraint(
            data_ID1="second",
            data_ID2="first",
        )
        is None
    )


# ==============================================================================
# test_BinaryConstraintsManager_check_transitivity_after_constraint_deletion
# ==============================================================================
def test_BinaryConstraintsManager_check_transitivity_after_constraint_deletion():
    """
    Test that the `delete_constraint` method of the `constraints.binary.BinaryConstraintsManager` class is transitive.
    """

    # Initialize a classic binaray constraints manager.
    constraints_manager = BinaryConstraintsManager(
        list_of_data_IDs=["first", "second", "third", "fourth"],
    )

    # Add `"MUST_LINK"` constraint between `"first"` and `"second"`.
    constraints_manager.add_constraint(
        data_ID1="first",
        data_ID2="second",
        constraint_type="MUST_LINK",
    )

    # Add `"MUST_LINK"` constraint between `"second"` and `"third"`.
    constraints_manager.add_constraint(
        data_ID1="second",
        data_ID2="third",
        constraint_type="MUST_LINK",
    )
    assert (
        constraints_manager.get_inferred_constraint(
            data_ID1="first",
            data_ID2="third",
        )
        == "MUST_LINK"
    )
    assert (
        constraints_manager.get_inferred_constraint(
            data_ID1="third",
            data_ID2="first",
        )
        == "MUST_LINK"
    )

    # Add `"CANNOT_LINK"` constraint between `"second"` and `"fourth"`.
    constraints_manager.add_constraint(
        data_ID1="second",
        data_ID2="fourth",
        constraint_type="CANNOT_LINK",
    )
    assert (
        constraints_manager.get_inferred_constraint(
            data_ID1="first",
            data_ID2="fourth",
        )
        == "CANNOT_LINK"
    )
    assert (
        constraints_manager.get_inferred_constraint(
            data_ID1="fourth",
            data_ID2="first",
        )
        == "CANNOT_LINK"
    )

    # Delete constraint between `"second"` and `"first"`.
    constraints_manager.delete_constraint(
        data_ID1="second",
        data_ID2="first",
    )

    # Run assertion (no more transitivity).
    assert (
        constraints_manager.get_inferred_constraint(
            data_ID1="first",
            data_ID2="third",
        )
        is None
    )
    assert (
        constraints_manager.get_inferred_constraint(
            data_ID1="third",
            data_ID2="first",
        )
        is None
    )
    assert (
        constraints_manager.get_inferred_constraint(
            data_ID1="first",
            data_ID2="fourth",
        )
        is None
    )
    assert (
        constraints_manager.get_inferred_constraint(
            data_ID1="fourth",
            data_ID2="first",
        )
        is None
    )


# ==============================================================================
# test_BinaryConstraintsManager_check_transitivity_after_data_ID_deletion
# ==============================================================================
def test_BinaryConstraintsManager_check_transitivity_after_data_ID_deletion():
    """
    Test that the `delete_data_ID` method of the `constraints.binary.BinaryConstraintsManager` class is transitive.
    """

    # Initialize a classic binaray constraints manager.
    constraints_manager = BinaryConstraintsManager(
        list_of_data_IDs=["first", "second", "third", "fourth"],
    )

    # Add `"MUST_LINK"` constraint between `"first"` and `"second"`.
    constraints_manager.add_constraint(
        data_ID1="first",
        data_ID2="second",
        constraint_type="MUST_LINK",
    )
    # Add `"MUST_LINK"` constraint between `"second"` and `"third"`.
    constraints_manager.add_constraint(
        data_ID1="second",
        data_ID2="third",
        constraint_type="MUST_LINK",
    )
    assert (
        constraints_manager.get_inferred_constraint(
            data_ID1="first",
            data_ID2="third",
        )
        == "MUST_LINK"
    )
    assert (
        constraints_manager.get_inferred_constraint(
            data_ID1="third",
            data_ID2="first",
        )
        == "MUST_LINK"
    )

    # Add `"CANNOT_LINK"` constraint between `"second"` and `"fourth"`.
    constraints_manager.add_constraint(
        data_ID1="second",
        data_ID2="fourth",
        constraint_type="CANNOT_LINK",
    )
    assert (
        constraints_manager.get_inferred_constraint(
            data_ID1="first",
            data_ID2="fourth",
        )
        == "CANNOT_LINK"
    )
    assert (
        constraints_manager.get_inferred_constraint(
            data_ID1="fourth",
            data_ID2="first",
        )
        == "CANNOT_LINK"
    )

    # Delete data ID `"second"`.
    constraints_manager.delete_data_ID(
        data_ID="second",
    )

    # Run assertion (no more transitivity).
    assert (
        constraints_manager.get_inferred_constraint(
            data_ID1="first",
            data_ID2="third",
        )
        is None
    )
    assert (
        constraints_manager.get_inferred_constraint(
            data_ID1="third",
            data_ID2="first",
        )
        is None
    )
    assert (
        constraints_manager.get_inferred_constraint(
            data_ID1="first",
            data_ID2="fourth",
        )
        is None
    )
    assert (
        constraints_manager.get_inferred_constraint(
            data_ID1="fourth",
            data_ID2="first",
        )
        is None
    )


# ==============================================================================
# test_BinaryConstraintsManager_get_list_of_involved_data_IDs_in_a_constraint_conflict_with_incorrect_data_ID
# ==============================================================================
def test_BinaryConstraintsManager_get_list_of_involved_data_IDs_in_a_constraint_conflict_with_incorrect_data_ID():
    """
    Test that the `get_list_of_involved_data_IDs_in_a_constraint_conflict` method of the `constraints.binary.BinaryConstraintsManager` class raises `ValueError` for incorrect data IDs.
    """

    # Initialize a classic binary constraints manager.
    constraints_manager = BinaryConstraintsManager(
        list_of_data_IDs=["first", "second", "third"],
    )

    # Try with `"unknown"` data ID in `"data_ID1"`.
    with pytest.raises(ValueError, match="`data_ID1`"):
        constraints_manager.get_list_of_involved_data_IDs_in_a_constraint_conflict(
            data_ID1="unknown",
            data_ID2="second",
            constraint_type="MUST_LINK",
        )

    # Try with `"unknown"` data ID in `"data_ID2"`.
    with pytest.raises(ValueError, match="`data_ID2`"):
        constraints_manager.get_list_of_involved_data_IDs_in_a_constraint_conflict(
            data_ID1="first",
            data_ID2="unknown",
            constraint_type="CANNOT_LINK",
        )


# ==============================================================================
# test_BinaryConstraintsManager_get_list_of_involved_data_IDs_in_a_constraint_conflict_with_incorrect_constraint_type
# ==============================================================================
def test_BinaryConstraintsManager_get_list_of_involved_data_IDs_in_a_constraint_conflict_with_incorrect_constraint_type():
    """
    Test that the `get_list_of_involved_data_IDs_in_a_constraint_conflict` method of the `constraints.binary.BinaryConstraintsManager` class raises `ValueError` for an incorrect constraint type.
    """

    # Initialize a classic binary constraints manager.
    constraints_manager = BinaryConstraintsManager(
        list_of_data_IDs=["first", "second", "third"],
    )

    # Try with `"UNKNOWN_LINK"` constraint type in `"constraint_type"`.
    with pytest.raises(ValueError, match="`constraint_type`"):
        constraints_manager.get_list_of_involved_data_IDs_in_a_constraint_conflict(
            data_ID1="first",
            data_ID2="second",
            constraint_type="UNKNOWN_LINK",
        )


# ==============================================================================
# test_BinaryConstraintsManager_get_list_of_involved_data_IDs_in_a_constraint_conflict
# ==============================================================================
def test_BinaryConstraintsManager_get_list_of_involved_data_IDs_in_a_constraint_conflict():
    """
    Test that the `get_list_of_involved_data_IDs_in_a_constraint_conflict` method of the `constraints.binary.BinaryConstraintsManager` class works.
    """

    # Initialize a classic binary constraints manager.
    constraints_manager = BinaryConstraintsManager(
        list_of_data_IDs=["first", "second", "third", "fourth", "fifth"],
    )
    constraints_manager.add_constraint(
        data_ID1="first",
        data_ID2="second",
        constraint_type="MUST_LINK",
    )
    constraints_manager.add_constraint(
        data_ID1="second",
        data_ID2="third",
        constraint_type="MUST_LINK",
    )
    constraints_manager.add_constraint(
        data_ID1="third",
        data_ID2="fourth",
        constraint_type="CANNOT_LINK",
    )
    constraints_manager.add_constraint(
        data_ID1="fourth",
        data_ID2="fifth",
        constraint_type="CANNOT_LINK",
    )

    # Conflict of adding `"MUST_LINK"` between "first" and "fourth".
    assert (
        constraints_manager.get_list_of_involved_data_IDs_in_a_constraint_conflict(
            data_ID1="first",
            data_ID2="fourth",
            constraint_type="MUST_LINK",
        )
        == ["first", "second", "third", "fourth"]
    )

    # Conflict of adding `"CANNOT_LINK"` between "first" and "fourth".
    assert (
        constraints_manager.get_list_of_involved_data_IDs_in_a_constraint_conflict(
            data_ID1="first",
            data_ID2="third",
            constraint_type="CANNOT_LINK",
        )
        == ["first", "second", "third"]
    )

    # No conflict.
    assert (
        constraints_manager.get_list_of_involved_data_IDs_in_a_constraint_conflict(
            data_ID1="first",
            data_ID2="third",
            constraint_type="MUST_LINK",
        )
        is None
    )

    # No conflict.
    assert (
        constraints_manager.get_list_of_involved_data_IDs_in_a_constraint_conflict(
            data_ID1="fourth",
            data_ID2="fifth",
            constraint_type="CANNOT_LINK",
        )
        is None
    )



# -*- coding: utf-8 -*-

"""
* Name:         cognitivefactory.interactive_clustering.utils.vectorization
* Description:  Utilities methods to apply NLP vectorization.
* Author:       Erwan Schild
* Created:      17/03/2021
* Licence:      CeCILL (https://cecill.info/licences.fr.html)
"""

# ==============================================================================
# IMPORT PYTHON DEPENDENCIES
# ==============================================================================

from typing import Dict, Union  # To type Python code (mypy).

import numpy as np  # To handle float.
import spacy  # To apply spacy language models.
from numpy import ndarray  # To handle matrix and vectors.
from scipy.sparse import csr_matrix  # To handle matrix and vectors.
from sklearn.feature_extraction.text import TfidfVectorizer  # To apply TF-IDF vectorisation.


# ==============================================================================
# NLP VECTORIZATION
# ==============================================================================
def vectorize(
    dict_of_texts: Dict[str, str],
    vectorizer_type: str = "tfidf",
    spacy_language_model: str = "fr_core_news_sm",
) -> Dict[str, Union[ndarray, csr_matrix]]:
    """
    A method used to vectorize texts.
    Severals vectorizer are available : TFIDF, spaCy language model.

    References:
        - _Scikit-learn_: `Pedregosa, F., G. Varoquaux, A. Gramfort, V. Michel, B. Thirion, O. Grisel, M. Blondel, P. Prettenhofer, R.Weiss, V. Dubourg, J. Vanderplas, A. Passos, D. Cournapeau, M. Brucher, M. Perrot, et E. Duchesnay (2011). Scikit-learn : Machine Learning in Python. Journal of Machine Learning Research 12, 2825–2830.`
        - _Scikit-learn_ _'TfidfVectorizer'_: `https://scikit-learn.org/stable/modules/generated/sklearn.feature_extraction.text.TfidfVectorizer.html`
        - _spaCy_: `Honnibal, M. et I. Montani (2017). spaCy 2 : Natural language understanding with Bloom embeddings, convolutional neural networks and incremental parsing.`
        - _spaCy_ language models: `https://spacy.io/usage/models`

    Args:
        dict_of_texts (Dict[str,str]): A dictionary that contains the texts to vectorize.
        vectorizer_type (str, optional): The vectorizer type to use. The type can be `"tfidf"` or `"spacy"`. Defaults to `"tfidf"`.
        spacy_language_model (str, optional): The spaCy language model to use if vectorizer is spacy. Defaults to `"fr_core_news_sm"`.

    Raises:
        ValueError: Raises error if `vectorizer_type` is not implemented or if the `spacy_language_model` is not installed.

    Returns:
        Dict[str, Union[ndarray,csr_matrix]]: A dictionary that contains the computed vectors.

    Examples:
        ```python
        # Import.
        from cognitivefactory.interactive_clustering.utils.vectorization import vectorize

        # Define data.
        dict_of_texts={
            "0": "comment signaler une perte de carte de paiement",
            "1": "quelle est la procedure pour chercher une carte de credit avalee",
            "2": "ma carte visa a un plafond de paiment trop bas puis je l augmenter",
        }

        # Apply vectorization.
        dict_of_vectors = vectorize(
            dict_of_texts=dict_of_texts
            vectorizer_type="spacy",
            spacy_language_model="fr_core_news_sm",
        )

        # Print results.
        print("Computed results", ":", dict_of_vectors)
        ```
    """

    # Initialize dictionary of vectors.
    dict_of_vectors: Dict[str, Union[ndarray, csr_matrix]] = {}

    ###
    ### Case of TFIDF vectorization.
    ###
    if vectorizer_type == "tfidf":

        # Initialize vectorizer.
        vectorizer = TfidfVectorizer(
            analyzer="word",
            ngram_range=(1, 3),
            min_df=2,
        )

        # Apply vectorization.
        tfidf_vectorization: csr_matrix = vectorizer.fit_transform(
            [str(dict_of_texts[data_ID]) for data_ID in dict_of_texts.keys()]
        )

        # Format dictionary of vectors to return.
        dict_of_vectors = {
            data_ID: tfidf_vectorization[i].astype(np.float64) for i, data_ID in enumerate(dict_of_texts.keys())
        }

        # Return the dictionary of vectors.
        return dict_of_vectors

    ###
    ### Case of SPACY vectorization.
    ###
    if vectorizer_type == "spacy":

        # Load vectorizer (spaCy language model).
        try:
            spacy_nlp = spacy.load(
                name=spacy_language_model,
                disable=[
                    "tagger",  # Not needed
                    "parser",  # Not needed
                    "ner",  # Not needed
                ],
            )
        except OSError as err:  # `spacy_language_model` is not installed.
            raise ValueError(
                "The `spacy_language_model` '" + str(spacy_language_model) + "' is not installed."
            ) from err

        # Apply vectorization.
        dict_of_vectors = {
            data_ID: spacy_nlp(str(text)).vector.astype(np.float64) for data_ID, text in dict_of_texts.items()
        }

        # Return the dictionary of vectors.
        return dict_of_vectors

    ###
    ### Other case : Raise a `ValueError`.
    ###
    raise ValueError("The `vectorizer_type` '" + str(vectorizer_type) + "' is not implemented.")



# -*- coding: utf-8 -*-

"""
* Name:         interactive-clustering/tests/utils/test_preprocessing.py
* Description:  Unittests for the `utils.preprocessing` module.
* Author:       Erwan Schild
* Created:      17/03/2021
* Licence:      CeCILL (https://cecill.info/licences.fr.html)
"""

# ==============================================================================
# IMPORT PYTHON DEPENDENCIES
# ==============================================================================

import pytest

from cognitivefactory.interactive_clustering.utils.preprocessing import preprocess


# ==============================================================================
# test_preprocess_for_uninstalled_spacy_language_model
# ==============================================================================
def test_preprocess_for_uninstalled_spacy_language_model():
    """
    Test that the `utils.preprocessing.preprocess` raises `ValueError` for uninstalled spacy language model.
    """

    # Check a unimplemented vectorizer.
    with pytest.raises(ValueError, match="`spacy_language_model`"):

        preprocess(
            dict_of_texts={
                "0": "Comment signaler une perte de carte de paiement ?",
                "1": "Quelle est la procédure pour chercher une carte de crédit avalée ?",
                "2": "Ma carte Visa a un plafond de paiment trop bas, puis-je l'augmenter ?",
            },
            spacy_language_model="uninstalled",
        )


# ==============================================================================
# test_preprocess_for_installed_spacy_language_model
# ==============================================================================
def test_preprocess_for_installed_spacy_language_model():
    """
    Test that the `utils.preprocessing.preprocess` works for an installed spacy language model.
    """

    # Check simple preprocessing.
    dict_of_preprocessed_texts = preprocess(
        dict_of_texts={
            "0": "Hello. How are you ??",
            "1": "Hello, how old are you ?",
            "2": "Hello ! Where do you live ?",
        },
        spacy_language_model="en_core_web_sm",
    )

    # Assertions
    assert dict_of_preprocessed_texts


# ==============================================================================
# test_preprocess_for_simple_preprocessing
# ==============================================================================
def test_preprocess_for_simple_preprocessing():
    """
    Test that the `utils.preprocessing.preprocess` works for simple preprocessing.
    """

    # Check simple preprocessing.
    dict_of_preprocessed_texts = preprocess(
        dict_of_texts={
            "0": "Comment signaler une perte de carte de paiement ?",
            "1": "Quelle est la procédure pour chercher une carte de crédit avalée ?",
            "2": "Ma carte Visa a un plafond de paiment trop bas, puis-je l'augmenter ?",
        },
    )

    # Assertions
    assert dict_of_preprocessed_texts
    assert sorted(dict_of_preprocessed_texts.keys()) == ["0", "1", "2"]
    assert dict_of_preprocessed_texts["0"] == "comment signaler une perte de carte de paiement"
    assert dict_of_preprocessed_texts["1"] == "quelle est la procedure pour chercher une carte de credit avalee"
    assert dict_of_preprocessed_texts["2"] == "ma carte visa a un plafond de paiment trop bas puis je l augmenter"


# ==============================================================================
# test_preprocess_for_stopwords_deletion
# ==============================================================================
def test_preprocess_for_stopwords_deletion():
    """
    Test that the `utils.preprocessing.preprocess` works for stopwords deletion.
    """

    # Check stopwords deletion.
    dict_of_preprocessed_texts = preprocess(
        dict_of_texts={
            "0": "Comment signaler une perte de carte de paiement ?",
            "1": "Quelle est la procédure pour chercher une carte de crédit avalée ?",
            "2": "Ma carte Visa a un plafond de paiment trop bas, puis-je l'augmenter ?",
        },
        apply_stopwords_deletion=True,
    )

    # Assertions
    assert dict_of_preprocessed_texts
    assert sorted(dict_of_preprocessed_texts.keys()) == ["0", "1", "2"]
    assert dict_of_preprocessed_texts["0"] == "signaler perte carte paiement"
    assert dict_of_preprocessed_texts["1"] == "procedure chercher carte credit avalee"
    assert dict_of_preprocessed_texts["2"] == "carte visa plafond paiment l augmenter"


# ==============================================================================
# test_preprocess_for_parsing_filter
# ==============================================================================
def test_preprocess_for_parsing_filter():
    """
    Test that the `utils.preprocessing.preprocess` works for parsing filter.
    """

    # Check parsing filter.
    dict_of_preprocessed_texts = preprocess(
        dict_of_texts={
            "0": "Comment signaler une perte de carte de paiement ?",
            "1": "Quelle est la procédure pour chercher une carte de crédit avalée ?",
            "2": "Ma carte Visa a un plafond de paiment trop bas, puis-je l'augmenter ?",
        },
        apply_parsing_filter=True,
    )

    # Assertions
    assert dict_of_preprocessed_texts
    assert sorted(dict_of_preprocessed_texts.keys()) == ["0", "1", "2"]
    assert dict_of_preprocessed_texts["0"] == "comment signaler perte"
    assert dict_of_preprocessed_texts["1"] == "quelle est la procedure chercher"
    assert dict_of_preprocessed_texts["2"] == "carte a plafond l"


# ==============================================================================
# test_preprocess_for_lemmatization
# ==============================================================================
def test_preprocess_for_lemmatization():
    """
    Test that the `utils.preprocessing.preprocess` works for lemmatization.
    """

    # Check lemmatization.
    dict_of_preprocessed_texts = preprocess(
        dict_of_texts={
            "0": "Comment signaler une perte de carte de paiement ?",
            "1": "Quelle est la procédure pour chercher une carte de crédit avalée ?",
            "2": "Ma carte Visa a un plafond de paiment trop bas, puis-je l'augmenter ?",
        },
        apply_lemmatization=True,
    )

    # Assertions
    assert dict_of_preprocessed_texts
    assert sorted(dict_of_preprocessed_texts.keys()) == ["0", "1", "2"]
    assert dict_of_preprocessed_texts["0"] == "comment signaler un perte de carte de paiement"
    assert dict_of_preprocessed_texts["1"] == "quell etre le procedure pour chercher un carte de credit avaler"
    assert dict_of_preprocessed_texts["2"] == "mon carte visa avoir un plafond de paiment trop bas puis je l augmenter"



# -*- coding: utf-8 -*-

"""
* Name:         interactive-clustering/tests/sampling/test_abstract.py
* Description:  Unittests for the `sampling.abstract` module.
* Author:       Erwan Schild
* Created:      17/03/2021
* Licence:      CeCILL (https://cecill.info/licences.fr.html)
"""

# ==============================================================================
# IMPORT PYTHON DEPENDENCIES
# ==============================================================================

import pytest

from cognitivefactory.interactive_clustering.sampling.abstract import AbstractConstraintsSampling


# ==============================================================================
# test_AbstractConstraintsSampling_is_abstract
# ==============================================================================
def test_AbstractConstraintsSampling_is_abstract():
    """
    Test that the `sampling.abstract.AbstractConstraintsSampling` class is abstract.
    """

    # Check `TypeError` for initialization of `AbstractConstraintsSampling`.
    with pytest.raises(TypeError, match="Can't instantiate abstract class"):
        AbstractConstraintsSampling()



# -*- coding: utf-8 -*-

"""
* Name:         interactive-clustering/tests/sampling/test_factory.py
* Description:  Unittests for the `sampling.factory` module.
* Author:       Erwan Schild
* Created:      17/03/2021
* Licence:      CeCILL (https://cecill.info/licences.fr.html)
"""

# ==============================================================================
# IMPORT PYTHON DEPENDENCIES
# ==============================================================================

import pytest

from cognitivefactory.interactive_clustering.sampling.clusters_based import ClustersBasedConstraintsSampling
from cognitivefactory.interactive_clustering.sampling.factory import sampling_factory


# ==============================================================================
# test_sampling_factory_for_not_implemented_clustering
# ==============================================================================
def test_sampling_factory_for_not_implemented_clustering():
    """
    Test that the `sampling.sampling_factory` method raises an `ValueError` for not implemented clustering.
    """

    # Check `ValueError` for bad string value for `algorithm`.
    with pytest.raises(ValueError, match="`algorithm`"):
        sampling_factory(
            algorithm="unknown",
        )


# ==============================================================================
# test_sampling_factory_for_random_sampling
# ==============================================================================
def test_sampling_factory_for_random_sampling():
    """
    Test that the `sampling.sampling_factory` can initialize an instance of `ClustersBasedConstraintsSampling` to sample random pairs of data IDs.
    """

    # Check average `random` sampling.
    sampling_model = sampling_factory(
        algorithm="random",
    )
    assert isinstance(sampling_model, ClustersBasedConstraintsSampling)
    assert sampling_model.clusters_restriction is None
    assert sampling_model.distance_restriction is None


# ==============================================================================
# test_sampling_factory_for_random_in_same_cluster_sampling
# ==============================================================================
def test_sampling_factory_for_random_in_same_cluster_sampling():
    """
    Test that the `sampling.sampling_factory` can initialize an instance of `ClustersBasedConstraintsSampling` to sample random pairs of data IDs in same clusters.
    """

    # Check average `random_in_same_cluster` sampling.
    sampling_model = sampling_factory(
        algorithm="random_in_same_cluster",
    )
    assert isinstance(sampling_model, ClustersBasedConstraintsSampling)
    assert sampling_model.clusters_restriction == "same_cluster"
    assert sampling_model.distance_restriction is None


# ==============================================================================
# test_sampling_factory_for_farthest_in_same_cluster_sampling
# ==============================================================================
def test_sampling_factory_for_farthest_in_same_cluster_sampling():
    """
    Test that the `sampling.sampling_factory` can initialize an instance of `ClustersBasedConstraintsSampling` to sample farthest pairs of data IDs in same clusters.
    """

    # Check average `farthest_in_same_cluster` sampling.
    sampling_model = sampling_factory(
        algorithm="farthest_in_same_cluster",
    )
    assert isinstance(sampling_model, ClustersBasedConstraintsSampling)
    assert sampling_model.clusters_restriction == "same_cluster"
    assert sampling_model.distance_restriction == "farthest_neighbors"


# ==============================================================================
# test_sampling_factory_for_closest_in_different_clusters_sampling
# ==============================================================================
def test_sampling_factory_for_closest_in_different_clusters_sampling():
    """
    Test that the `sampling.sampling_factory` can initialize an instance of `ClustersBasedConstraintsSampling` to sample closest pairs of data IDs in different clusters.
    """

    # Check average `closest_in_different_clusters` sampling.
    sampling_model = sampling_factory(
        algorithm="closest_in_different_clusters",
    )
    assert isinstance(sampling_model, ClustersBasedConstraintsSampling)
    assert sampling_model.clusters_restriction == "different_clusters"
    assert sampling_model.distance_restriction == "closest_neighbors"



# -*- coding: utf-8 -*-

"""
* Name:         interactive-clustering/tests/utils/test_frequency.py
* Description:  Unittests for the `utils.frequency` module.
* Author:       Erwan Schild
* Created:      17/03/2021
* Licence:      CeCILL (https://cecill.info/licences.fr.html)
"""

# ==============================================================================
# IMPORT PYTHON DEPENDENCIES
# ==============================================================================

from cognitivefactory.interactive_clustering.utils.frequency import compute_clusters_frequency


# ==============================================================================
# test_compute_clusters_frequency_for_no_clustering_result
# ==============================================================================
def test_compute_clusters_frequency_for_no_clustering_result():
    """
    Test that the `utils.frequency.compute_clusters_frequency` works for no `clustering_result`.
    """

    # Check for no `clustering_result`.
    assert not compute_clusters_frequency(clustering_result={})


# ==============================================================================
# test_compute_clusters_frequency_for_correct_clustering_result
# ==============================================================================
def test_compute_clusters_frequency_for_correct_clustering_result():
    """
    Test that the `utils.frequency.compute_clusters_frequency` works for correct `clustering_result`.
    """

    # Check for no `clustering_result`.
    assert compute_clusters_frequency(
        clustering_result={
            "01": 0,
            "02": 0,
            "03": 0,
            "04": 0,
            "05": 0,
            "06": 1,
            "07": 1,
            "08": 2,
            "09": 2,
            "10": 2,
        }
    ) == {
        0: 0.5,
        1: 0.2,
        2: 0.3,
    }



# -*- coding: utf-8 -*-

"""
* Name:         cognitivefactory.interactive_clustering
* Description:  Python package used to apply NLP interactive clustering methods.
* Author:       Erwan Schild
* Created:      17/03/2021
* Licence:      CeCILL (https://cecill.info/licences.fr.html)
"""

from typing import List

__all__: List[str] = []  # noqa: WPS410 (only __variable__ we use)



# -*- coding: utf-8 -*-

"""
* Name:         interactive-clustering/tests/clustering/test_spectral.py
* Description:  Unittests for the `clustering.spectral` module.
* Author:       Erwan Schild
* Created:      17/03/2021
* Licence:      CeCILL (https://cecill.info/licences.fr.html)
"""

# ==============================================================================
# IMPORT PYTHON DEPENDENCIES
# ==============================================================================

import numpy as np
import pytest
from scipy.sparse import csr_matrix

from cognitivefactory.interactive_clustering.clustering.spectral import SpectralConstrainedClustering
from cognitivefactory.interactive_clustering.constraints.binary import BinaryConstraintsManager


# ==============================================================================
# test_SpectralConstrainedClustering_for_inconsistent_model
# ==============================================================================
def test_SpectralConstrainedClustering_for_inconsistent_model():
    """
    Test that the `clustering.spectral.SpectralConstrainedClustering` initialization raises an `ValueError` for inconsistent `model` parameter.
    """

    # Check `ValueError` for bad string value for `model`.
    with pytest.raises(ValueError, match="`model`"):
        SpectralConstrainedClustering(
            model="as_you_want",
        )


# ==============================================================================
# test_SpectralConstrainedClustering_for_inconsistent_nb_components
# ==============================================================================
def test_SpectralConstrainedClustering_for_inconsistent_nb_components():
    """
    Test that the `clustering.spectral.SpectralConstrainedClustering` initialization raises an `ValueError` for inconsistent `nb_components` parameter.
    """

    # Check `ValueError` for bad string value for `nb_components`.
    with pytest.raises(ValueError, match="`nb_components`"):
        SpectralConstrainedClustering(
            nb_components=-1,
        )


# ==============================================================================
# test_SpectralConstrainedClustering_for_correct_settings
# ==============================================================================
def test_SpectralConstrainedClustering_for_correct_settings():
    """
    Test that the `clustering.spectral.SpectralConstrainedClustering` initialization runs correctly with the correct settings.
    """

    # Check a correct initialization.
    clustering_model = SpectralConstrainedClustering(
        model="SPEC",
        nb_components=100,
        random_seed=3,
    )
    assert clustering_model
    assert clustering_model.model == "SPEC"
    assert clustering_model.nb_components == 100
    assert clustering_model.random_seed == 3


# ==============================================================================
# test_SpectralConstrainedClustering_cluster_for_inconsistent_constraints_manager
# ==============================================================================
def test_SpectralConstrainedClustering_cluster_for_inconsistent_constraints_manager():
    """
    Test that the `clustering.spectral.SpectralConstrainedClustering` clustering raises an `ValueError` for inconsistent `constraints_manager` parameter.
    """

    # Initialize a `SpectralConstrainedClustering` instance.
    clustering_model = SpectralConstrainedClustering()

    # Check `ValueError` for not matrix `vectors`.
    with pytest.raises(ValueError, match="`constraints_manager`"):
        clustering_model.cluster(
            constraints_manager=None,
            vectors=None,
            nb_clusters=2,
        )


# ==============================================================================
# test_SpectralConstrainedClustering_cluster_for_inconsistent_vectors
# ==============================================================================
def test_SpectralConstrainedClustering_cluster_for_inconsistent_vectors():
    """
    Test that the `clustering.spectral.SpectralConstrainedClustering` clustering raises an `ValueError` for inconsistent `vectors` parameter.
    """

    # Initialize a `SpectralConstrainedClustering` instance.
    clustering_model = SpectralConstrainedClustering()

    # Check `ValueError` for not matrix `vectors`.
    with pytest.raises(ValueError, match="`vectors`"):
        clustering_model.cluster(
            constraints_manager=BinaryConstraintsManager(list_of_data_IDs=["first", "second", "third"]),
            vectors=None,
            nb_clusters=2,
        )


# ==============================================================================
# test_SpectralConstrainedClustering_cluster_for_inconsistent_nb_clusters
# ==============================================================================
def test_SpectralConstrainedClustering_cluster_for_inconsistent_nb_clusters():
    """
    Test that the `clustering.spectral.SpectralConstrainedClustering` clustering raises an `ValueError` for inconsistent `nb_clusters` parameter.
    """

    # Initialize a `SpectralConstrainedClustering` instance.
    clustering_model = SpectralConstrainedClustering()

    # Check `ValueError` for too small `nb_clusters`.
    with pytest.raises(ValueError, match="`nb_clusters`"):
        clustering_model.cluster(
            constraints_manager=BinaryConstraintsManager(list_of_data_IDs=["first", "second", "third"]),
            vectors={"first": np.array([1, 2, 3]), "second": np.array([[4, 5, 6]]), "third": csr_matrix([7, 8, 9])},
            nb_clusters=-1,
        )


# ==============================================================================
# test_SpectralConstrainedClustering_cluster_model_SPEC_with_no_constraints
# ==============================================================================
def test_SpectralConstrainedClustering_cluster_model_SPEC_with_no_constraints():
    """
    Test that the `clustering.spectral.SpectralConstrainedClustering` clustering works with SPEC `model` and no constraints.
    """

    # Define `vectors` and `constraints_manager`
    vectors = {
        "0": csr_matrix([1.00, 0.00, 0.00, 0.00]),
        "1": csr_matrix([0.95, 0.02, 0.02, 0.01]),
        "2": csr_matrix([0.98, 0.00, 0.02, 0.00]),
        "3": csr_matrix([0.99, 0.00, 0.01, 0.00]),
        "4": csr_matrix([0.60, 0.17, 0.16, 0.07]),
        "5": csr_matrix([0.60, 0.16, 0.17, 0.07]),
        "6": csr_matrix([0.01, 0.01, 0.01, 0.97]),
        "7": csr_matrix([0.00, 0.01, 0.00, 0.99]),
        "8": csr_matrix([0.00, 0.00, 0.00, 1.00]),
    }
    constraints_manager = BinaryConstraintsManager(list_of_data_IDs=list(vectors.keys()))

    # Initialize a `SpectralConstrainedClustering` instance.
    clustering_model = SpectralConstrainedClustering(
        model="SPEC",
        random_seed=1,
    )

    # Run clustering 3 clusters and no constraints.
    dict_of_predicted_clusters = clustering_model.cluster(
        constraints_manager=constraints_manager,
        vectors=vectors,
        nb_clusters=3,
    )

    assert clustering_model.dict_of_predicted_clusters
    assert dict_of_predicted_clusters == {
        "0": 0,
        "1": 0,
        "2": 0,
        "3": 0,
        "4": 1,
        "5": 1,
        "6": 2,
        "7": 2,
        "8": 2,
    }


# ==============================================================================
# test_SpectralConstrainedClustering_cluster_model_SPEC_with_some_constraints
# ==============================================================================
def test_SpectralConstrainedClustering_cluster_model_SPEC_with_some_constraints():
    """
    Test that the `clustering.spectral.SpectralConstrainedClustering` clustering works with SPEC `model` and some constraints.
    """

    # Define `vectors` and `constraints_manager`
    vectors = {
        "0": csr_matrix([1.00, 0.00, 0.00, 0.00]),
        "1": csr_matrix([0.95, 0.02, 0.02, 0.01]),
        "2": csr_matrix([0.98, 0.00, 0.02, 0.00]),
        "3": csr_matrix([0.99, 0.00, 0.01, 0.00]),
        "4": csr_matrix([0.60, 0.17, 0.16, 0.07]),
        "5": csr_matrix([0.60, 0.16, 0.17, 0.07]),
        "6": csr_matrix([0.01, 0.01, 0.01, 0.97]),
        "7": csr_matrix([0.00, 0.01, 0.00, 0.99]),
        "8": csr_matrix([0.00, 0.00, 0.00, 1.00]),
    }
    constraints_manager = BinaryConstraintsManager(list_of_data_IDs=["0", "1", "2", "3", "4", "5", "6", "7", "8"])
    constraints_manager.add_constraint(data_ID1="0", data_ID2="1", constraint_type="MUST_LINK")
    constraints_manager.add_constraint(data_ID1="2", data_ID2="3", constraint_type="MUST_LINK")
    constraints_manager.add_constraint(data_ID1="4", data_ID2="5", constraint_type="MUST_LINK")
    constraints_manager.add_constraint(data_ID1="7", data_ID2="8", constraint_type="MUST_LINK")
    constraints_manager.add_constraint(data_ID1="0", data_ID2="4", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="2", data_ID2="4", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="4", data_ID2="7", constraint_type="CANNOT_LINK")

    # Initialize a `SpectralConstrainedClustering` instance.
    clustering_model = SpectralConstrainedClustering(
        model="SPEC",
        random_seed=1,
    )

    # Run clustering 3 clusters and some constraints.
    dict_of_predicted_clusters = clustering_model.cluster(
        constraints_manager=constraints_manager,
        vectors=vectors,
        nb_clusters=3,
    )

    assert clustering_model.dict_of_predicted_clusters
    assert dict_of_predicted_clusters == {
        "0": 0,
        "1": 0,
        "2": 0,
        "3": 0,
        "4": 1,
        "5": 1,
        "6": 2,
        "7": 2,
        "8": 2,
    }


# ==============================================================================
# test_SpectralConstrainedClustering_cluster_model_SPEC_with_full_constraints
# ==============================================================================
def test_SpectralConstrainedClustering_cluster_model_SPEC_with_full_constraints():
    """
    Test that the `clustering.spectral.SpectralConstrainedClustering` clustering works with SPEC `model` and full constraints.
    """

    # Define `vectors` and `constraints_manager`
    vectors = {
        "0": csr_matrix([1.00, 0.00, 0.00, 0.00]),
        "1": csr_matrix([0.95, 0.02, 0.02, 0.01]),
        "2": csr_matrix([0.98, 0.00, 0.02, 0.00]),
        "3": csr_matrix([0.99, 0.00, 0.01, 0.00]),
        "4": csr_matrix([0.60, 0.17, 0.16, 0.07]),
        "5": csr_matrix([0.60, 0.16, 0.17, 0.07]),
        "6": csr_matrix([0.01, 0.01, 0.01, 0.97]),
        "7": csr_matrix([0.00, 0.01, 0.00, 0.99]),
        "8": csr_matrix([0.00, 0.00, 0.00, 1.00]),
    }
    constraints_manager = BinaryConstraintsManager(list_of_data_IDs=["0", "1", "2", "3", "4", "5", "6", "7", "8"])
    constraints_manager.add_constraint(data_ID1="0", data_ID2="4", constraint_type="MUST_LINK")
    constraints_manager.add_constraint(data_ID1="0", data_ID2="8", constraint_type="MUST_LINK")
    constraints_manager.add_constraint(data_ID1="1", data_ID2="5", constraint_type="MUST_LINK")
    constraints_manager.add_constraint(data_ID1="2", data_ID2="6", constraint_type="MUST_LINK")
    constraints_manager.add_constraint(data_ID1="3", data_ID2="7", constraint_type="MUST_LINK")
    constraints_manager.add_constraint(data_ID1="0", data_ID2="1", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="0", data_ID2="2", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="0", data_ID2="3", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="1", data_ID2="2", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="1", data_ID2="3", constraint_type="CANNOT_LINK")
    constraints_manager.add_constraint(data_ID1="2", data_ID2="3", constraint_type="CANNOT_LINK")

    # Initialize a `SpectralConstrainedClustering` instance.
    clustering_model = SpectralConstrainedClustering(
        model="SPEC",
        random_seed=1,
    )

    # Run clustering 3 clusters and full constraints.
    dict_of_predicted_clusters = clustering_model.cluster(
        constraints_manager=constraints_manager,
        vectors=vectors,
        nb_clusters=4,
    )

    assert clustering_model.dict_of_predicted_clusters
    assert dict_of_predicted_clusters == {
        "0": 0,
        "1": 1,
        "2": 2,
        "3": 3,
        "4": 0,
        "5": 1,
        "6": 2,
        "7": 3,
        "8": 0,
    }



# -*- coding: utf-8 -*-

"""
* Name:         cognitivefactory.interactive_clustering.sampling.abstract
* Description:  The abstract class used to define constraints sampling algorithms.
* Author:       Erwan Schild
* Created:      17/03/2021
* Licence:      CeCILL (https://cecill.info/licences.fr.html)
"""

# ==============================================================================
# IMPORT PYTHON DEPENDENCIES
# ==============================================================================

from abc import ABC, abstractmethod  # To define an abstract class.
from typing import Dict, List, Optional, Tuple, Union  # To type Python code (mypy).

from numpy import ndarray  # To handle matrix and vectors.
from scipy.sparse import csr_matrix  # To handle matrix and vectors.

from cognitivefactory.interactive_clustering.constraints.abstract import (  # To manage constraints.
    AbstractConstraintsManager,
)


# ==============================================================================
# ABSTRACT CONSTRAINTS SAMPLING
# ==============================================================================
class AbstractConstraintsSampling(ABC):
    """
    Abstract class that is used to define constraints sampling algorithms.
    The main inherited method is `sample`.
    """

    # ==============================================================================
    # ABSTRACT METHOD - SAMPLE
    # ==============================================================================
    @abstractmethod
    def sample(
        self,
        constraints_manager: AbstractConstraintsManager,
        nb_to_select: int,
        clustering_result: Optional[Dict[str, int]] = None,
        vectors: Optional[Dict[str, Union[ndarray, csr_matrix]]] = None,
        **kargs,
    ) -> List[Tuple[str, str]]:
        """
        (ABSTRACT METHOD)
        An abstract method that represents the main method used to sample couple of data IDs for constraints annotation.

        Args:
            constraints_manager (AbstractConstraintsManager): A constraints manager over data IDs.
            nb_to_select (int): The number of couple of data IDs to select.
            clustering_result (Optional[Dict[str,int]], optional): A dictionary that represents the predicted cluster for each data ID. The keys of the dictionary represents the data IDs. If `None`, no clustering result are used during the sampling. Defaults to `None`.
            vectors (Optional[Dict[str,Union[ndarray,csr_matrix]]], optional): vectors (Dict[str,Union[ndarray,csr_matrix]]): The representation of data vectors. The keys of the dictionary represents the data IDs. This keys have to refer to the list of data IDs managed by the `constraints_manager`. The value of the dictionary represent the vector of each data. Vectors can be dense (`numpy.ndarray`) or sparse (`scipy.sparse.csr_matrix`). If `None`, no vectors are used during the sampling. Defaults to `None`
            **kargs (dict): Other parameters that can be used in the sampling.

        Raises:
            ValueError: if some parameters are incorrectly set or incompatible.

        Returns:
            List[Tuple[str,str]]: A list of couple of data IDs.
        """



# -*- coding: utf-8 -*-

"""
* Name:         cognitivefactory.interactive_clustering.constraints.binary
* Description:  Implementation of binary constraints manager.
* Author:       Erwan Schild
* Created:      17/03/2021
* Licence:      CeCILL-C License v1.0 (https://cecill.info/licences.fr.html)
"""

# ==============================================================================
# IMPORT PYTHON DEPENDENCIES
# ==============================================================================

from typing import Dict, List, Optional, Set, Tuple  # To type Python code (mypy).

import networkx as nx  # To create graph.

from cognitivefactory.interactive_clustering.constraints.abstract import (  # To use abstract interface.
    AbstractConstraintsManager,
)


# ==============================================================================
# BINARY CONSTRAINTS MANAGER
# ==============================================================================
class BinaryConstraintsManager(AbstractConstraintsManager):
    """
    This class implements the binary constraints mangement.
    It inherits from `AbstractConstraintsManager`, and it takes into account the strong transitivity of constraints.

    References:
        - Binary constraints in clustering: `Wagstaff, K. et C. Cardie (2000). Clustering with Instance-level Constraints. Proceedings of the Seventeenth International Conference on Machine Learning, 1103–1110.`

    Examples:
        ```python
        # Import.
        from cognitivefactory.interactive_clustering.constraints.binary import BinaryConstraintsManager

        # Create an instance of binary constraints manager.
        constraints_manager = BinaryConstraintsManager(list_of_data_IDs=["0", "1", "2", "3", "4"])

        # Add new data ID.
        constraints_manager.add_data_ID(data_ID="99")

        # Get list of data IDs.
        constraints_manager.get_list_of_managed_data_IDs()

        # Delete an existing data ID.
        constraints_manager.delete_data_ID(data_ID="99")

        # Add constraints.
        constraints_manager.add_constraint(data_ID1="0", data_ID2="1", constraint_type="MUST_LINK")
        constraints_manager.add_constraint(data_ID1="1", data_ID2="2", constraint_type="MUST_LINK")
        constraints_manager.add_constraint(data_ID1="2, data_ID2="3", constraint_type="CANNOT_LINK")

        # Get added constraint.
        constraints_manager.get_added_constraint(data_ID1="0", data_ID2="1")  # expected ("MUST_LINK", 1.0)
        constraints_manager.get_added_constraint(data_ID1="0", data_ID2="2")  # expected None

        # Get inferred constraint.
        constraints_manager.get_inferred_constraint(data_ID1="0", data_ID2="2")  # expected "MUST_LINK"
        constraints_manager.get_inferred_constraint(data_ID1="0", data_ID2="3")  # expected "CANNOT_LINK"
        constraints_manager.get_inferred_constraint(data_ID1="0", data_ID2="4")  # expected None
        ```
    """

    # ==============================================================================
    # INITIALIZATION
    # ==============================================================================
    def __init__(self, list_of_data_IDs: List[str], **kargs) -> None:

        """
        The constructor for Binary Constraints Manager class.
        This class use the strong transitivity to infer on constraints, so constraints values are not taken into account.

        Args:
            list_of_data_IDs (List[str]): The list of data IDs to manage.
            **kargs (dict): Other parameters that can be used in the instantiation.

        Raises:
            ValueError: if `list_of_data_IDs` has duplicates.
        """

        # Define `self._allowed_constraint_types`.
        self._allowed_constraint_types: Set[str] = {
            "MUST_LINK",
            "CANNOT_LINK",
        }
        # Define `self._allowed_constraint_value_range`.
        self._allowed_constraint_value_range: Dict[str, float] = {
            "min": 1.0,
            "max": 1.0,
        }

        # If `list_of_data_IDs` has duplicates, raise `ValueError`.
        if len(list_of_data_IDs) != len(set(list_of_data_IDs)):
            raise ValueError("There is duplicates in `list_of_data_IDs`.")

        # Store `self.kargs` for binary constraints managing.
        self.kargs = kargs

        # Initialize `self._constraints_dictionary`.
        self._constraints_dictionary: Dict[str, Dict[str, Optional[Tuple[str, float]]]] = {
            data_ID1: {
                data_ID2: (
                    ("MUST_LINK", 1.0)
                    if (data_ID1 == data_ID2)
                    else None  # Unknwon constraints if `data_ID1` != `data_ID2`.
                )
                for data_ID2 in list_of_data_IDs
            }
            for data_ID1 in list_of_data_IDs
        }

        # Define `self._constraints_transitivity`.
        self._generate_constraints_transitivity()

    # ==============================================================================
    # DATA_ID MANAGEMENT - ADDITION
    # ==============================================================================
    def add_data_ID(
        self,
        data_ID: str,
    ) -> bool:

        """
        The main method used to add a new data ID to manage.

        Args:
            data_ID (str): The data ID to manage.

        Raises:
            ValueError: if `data_ID` is already managed.

        Returns:
            bool: `True` if the addition is done.
        """

        # If `data_ID` is in the data IDs that are currently managed, then raises a `ValueError`.
        if data_ID in self._constraints_dictionary.keys():
            raise ValueError("The `data_ID` `'" + str(data_ID) + "'` is already managed.")

        # Add `data_ID` to `self._constraints_dictionary.keys()`.
        self._constraints_dictionary[data_ID] = {}

        # Define constraint for `data_ID` and all other data IDs.
        for other_data_ID in self._constraints_dictionary.keys():
            self._constraints_dictionary[data_ID][other_data_ID] = (
                ("MUST_LINK", 1.0)
                if (data_ID == other_data_ID)
                else None  # Unknwon constraints if `data_ID` != `other_data_ID`.
            )
            self._constraints_dictionary[other_data_ID][data_ID] = (
                ("MUST_LINK", 1.0)
                if (data_ID == other_data_ID)
                else None  # Unknwon constraints if `data_ID1` != `other_data_ID`.
            )

        # Regenerate `self._constraints_transitivity`.
        # `Equivalent to `self._generate_constraints_transitivity()`
        self._constraints_transitivity[data_ID] = {
            "MUST_LINK": {data_ID: None},
            "CANNOT_LINK": {},
        }

        # Return `True`.
        return True

    # ==============================================================================
    # DATA_ID MANAGEMENT - DELETION
    # ==============================================================================
    def delete_data_ID(
        self,
        data_ID: str,
    ) -> bool:
        """
        The main method used to delete a data ID to no longer manage.

        Args:
            data_ID (str): The data ID to no longer manage.

        Raises:
            ValueError: if `data_ID` is not managed.

        Returns:
            bool: `True` if the deletion is done.
        """

        # If `data_ID` is not in the data IDs that are currently managed, then raises a `ValueError`.
        if data_ID not in self._constraints_dictionary.keys():
            raise ValueError("The `data_ID` `'" + str(data_ID) + "'` is not managed.")

        # Remove `data_ID` from `self._constraints_dictionary.keys()`.
        self._constraints_dictionary.pop(data_ID)

        # Remove `data_ID` from all `self._constraints_dictionary[other_data_ID].keys()`.
        for other_data_ID in self._constraints_dictionary.keys():
            self._constraints_dictionary[other_data_ID].pop(data_ID)

        # Regenerate `self._constraints_transitivity`
        self._generate_constraints_transitivity()

        # Return `True`.
        return True

    # ==============================================================================
    # DATA_ID MANAGEMENT - LISTING
    # ==============================================================================
    def get_list_of_managed_data_IDs(
        self,
    ) -> List[str]:
        """
        The main method used to get the list of data IDs that are managed.

        Returns:
            List[str]: The list of data IDs that are managed.
        """

        # Return the possible keys of `self._constraints_dictionary`.
        return list(self._constraints_dictionary.keys())

    # ==============================================================================
    # CONSTRAINTS MANAGEMENT - ADDITION
    # ==============================================================================
    def add_constraint(
        self,
        data_ID1: str,
        data_ID2: str,
        constraint_type: str,
        constraint_value: float = 1.0,
    ) -> bool:
        """
        The main method used to add a constraint between two data IDs.

        Args:
            data_ID1 (str): The first data ID that is concerned for this constraint addition.
            data_ID2 (str): The second data ID that is concerned for this constraint addition.
            constraint_type (str): The type of the constraint to add. The type have to be `"MUST_LINK"` or `"CANNOT_LINK"`.
            constraint_value (float, optional): The value of the constraint to add. The value have to be in range `[0.0, 1.0]`. Defaults to `1.0`.

        Raises:
            ValueError: if `data_ID1`, `data_ID2`, `constraint_type` are not managed, or if a conflict is detected with constraints inference.

        Returns:
            bool: `True` if the addition is done, `False` is the constraint can't be added.
        """

        # If `data_ID1` is not in the data IDs that are currently managed, then raises a `ValueError`.
        if data_ID1 not in self._constraints_dictionary.keys():
            raise ValueError("The `data_ID1` `'" + str(data_ID1) + "'` is not managed.")

        # If `data_ID2` is not in the data IDs that are currently managed, then raises a `ValueError`.
        if data_ID2 not in self._constraints_dictionary.keys():
            raise ValueError("The `data_ID2` `'" + str(data_ID2) + "'` is not managed.")

        # If the `constraint_type` is not in `self._allowed_constraint_types`, then raises a `ValueError`.
        if constraint_type not in self._allowed_constraint_types:
            raise ValueError(
                "The `constraint_type` `'"
                + str(constraint_type)
                + "'` is not managed. Allowed constraints types are : `"
                + str(self._allowed_constraint_types)
                + "`."
            )

        # Get current added constraint between `data_ID1` and `data_ID2`.
        inferred_constraint: Optional[str] = self.get_inferred_constraint(
            data_ID1=data_ID1,
            data_ID2=data_ID2,
        )

        # Case of conflict with constraints inference.
        if (inferred_constraint is not None) and (inferred_constraint != constraint_type):
            raise ValueError(
                "The `constraint_type` `'"
                + str(constraint_type)
                + "'` is incompatible with the inferred constraint `'"
                + str(inferred_constraint)
                + "'` between data IDs `'"
                + data_ID1
                + "'` and `'"
                + data_ID2
                + "'`."
            )

        # Get current added constraint between `data_ID1` and `data_ID2`.
        added_constraint: Optional[Tuple[str, float]] = self.get_added_constraint(
            data_ID1=data_ID1,
            data_ID2=data_ID2,
        )

        # If the constraint has already be added, ...
        if added_constraint is not None:
            # ... do nothing.
            return True  # `added_constraint[0] == constraint_type`.
        # Otherwise, the constraint has to be added.

        # Add the direct constraint between `data_ID1` and `data_ID2`.
        self._constraints_dictionary[data_ID1][data_ID2] = (constraint_type, 1.0)
        self._constraints_dictionary[data_ID2][data_ID1] = (constraint_type, 1.0)

        # Add the transitivity constraint between `data_ID1` and `data_ID2`.
        self._add_constraint_transitivity(
            data_ID1=data_ID1,
            data_ID2=data_ID2,
            constraint_type=constraint_type,
        )

        return True

    # ==============================================================================
    # CONSTRAINTS MANAGEMENT - DELETION
    # ==============================================================================
    def delete_constraint(
        self,
        data_ID1: str,
        data_ID2: str,
    ) -> bool:
        """
        The main method used to delete a constraint between two data IDs.

        Args:
            data_ID1 (str): The first data ID that is concerned for this constraint deletion.
            data_ID2 (str): The second data ID that is concerned for this constraint deletion.

        Raises:
            ValueError: if `data_ID1` or `data_ID2` are not managed.

        Returns:
            bool: `True` if the deletion is done, `False` if the constraint can't be deleted.
        """

        # If `data_ID1` is not in the data IDs that are currently managed, then raises a `ValueError`.
        if data_ID1 not in self._constraints_dictionary.keys():
            raise ValueError("The `data_ID1` `'" + str(data_ID1) + "'` is not managed.")

        # If `data_ID2` is not in the data IDs that are currently managed, then raises a `ValueError`.
        if data_ID2 not in self._constraints_dictionary.keys():
            raise ValueError("The `data_ID2` `'" + str(data_ID2) + "'` is not managed.")

        # Delete the constraint between `data_ID1` and `data_ID2`.
        self._constraints_dictionary[data_ID1][data_ID2] = None
        self._constraints_dictionary[data_ID2][data_ID1] = None

        # Regenerate `self._constraints_transitivity`.
        self._generate_constraints_transitivity()

        # Return `True`
        return True

    # ==============================================================================
    # CONSTRAINTS MANAGEMENT - GETTER
    # ==============================================================================
    def get_added_constraint(
        self,
        data_ID1: str,
        data_ID2: str,
    ) -> Optional[Tuple[str, float]]:
        """
        The main method used to get the constraint added between the two data IDs.
        Do not take into account the constraints transitivity, just look at constraints that are explicitly added.

        Args:
            data_ID1 (str): The first data ID that is concerned for this constraint.
            data_ID2 (str): The second data ID that is concerned for this constraint.

        Raises:
            ValueError: if `data_ID1` or `data_ID2` are not managed.

        Returns:
            Optional[Tuple[str, float]]: `None` if no constraint, `(constraint_type, constraint_value)` otherwise.
        """

        # If `data_ID1` is not in the data IDs that are currently managed, then raises a `ValueError`.
        if data_ID1 not in self._constraints_dictionary.keys():
            raise ValueError("The `data_ID1` `'" + str(data_ID1) + "'` is not managed.")

        # If `data_ID2` is not in the data IDs that are currently managed, then raises a `ValueError`.
        if data_ID2 not in self._constraints_dictionary.keys():
            raise ValueError("The `data_ID2` `'" + str(data_ID2) + "'` is not managed.")

        # Retrun the current added constraint type and value.
        return self._constraints_dictionary[data_ID1][data_ID2]

    # ==============================================================================
    # CONSTRAINTS EXPLORATION - GETTER
    # ==============================================================================
    def get_inferred_constraint(
        self,
        data_ID1: str,
        data_ID2: str,
        threshold: float = 1.0,
    ) -> Optional[str]:
        """
        The main method used to check if the constraint inferred by transitivity between the two data IDs.
        The transitivity is taken into account, and the `threshold` parameter is used to evaluate the impact of constraints transitivity.

        Args:
            data_ID1 (str): The first data ID that is concerned for this constraint.
            data_ID2 (str): The second data ID that is concerned for this constraint.
            threshold (float, optional): The threshold used to evaluate the impact of constraints transitivity link. Defaults to `1.0`.

        Raises:
            ValueError: if `data_ID1`, `data_ID2` or `threshold` are not managed.

        Returns:
            Optional[str]: The type of the inferred constraint. The type can be `None`, `"MUST_LINK"` or `"CANNOT_LINK"`.
        """

        # If `data_ID1` is not in the data IDs that are currently managed, then raises a `ValueError`.
        if data_ID1 not in self._constraints_transitivity.keys():
            raise ValueError("The `data_ID1` `'" + str(data_ID1) + "'` is not managed.")

        # If `data_ID2` is not in the data IDs that are currently managed, then raises a `ValueError`.
        if data_ID2 not in self._constraints_transitivity.keys():
            raise ValueError("The `data_ID2` `'" + str(data_ID2) + "'` is not managed.")

        # Case of `"MUST_LINK"`.
        if data_ID1 in self._constraints_transitivity[data_ID2]["MUST_LINK"].keys():
            return "MUST_LINK"

        # Case of `"CANNOT_LINK"`.
        if data_ID1 in self._constraints_transitivity[data_ID2]["CANNOT_LINK"].keys():
            return "CANNOT_LINK"

        # Case of `None`.
        return None

    # ==============================================================================
    # CONSTRAINTS EXPLORATION - LIST OF COMPONENTS GETTER
    # ==============================================================================
    def get_connected_components(
        self,
        threshold: float = 1.0,
    ) -> List[List[str]]:
        """
        The main method used to get the possible lists of data IDs that are linked by a `"MUST_LINK"` constraints.
        Each list forms a component of the constraints transitivity graph, and it forms a partition of the managed data IDs.
        The transitivity is taken into account, and the `threshold` parameters is used if constraints values are used in the constraints transitivity.

        Args:
            threshold (float, optional): The threshold used to define the transitivity link. Defaults to `1.0`.

        Returns:
            List[List[int]]: The list of lists of data IDs that represent a component of the constraints transitivity graph.
        """

        # Initialize the list of connected components.
        list_of_connected_components: List[List[str]] = []

        # For each data ID...
        for data_ID in self._constraints_transitivity.keys():

            # ... get the list of `"MUST_LINK"` data IDs linked by transitivity with `data_ID` ...
            connected_component_of_a_data_ID = list(self._constraints_transitivity[data_ID]["MUST_LINK"].keys())

            # ... and if the connected component is not already get...
            if connected_component_of_a_data_ID not in list_of_connected_components:
                # ... then add it to the list of connected components.
                list_of_connected_components.append(connected_component_of_a_data_ID)

        # Return the list of connected components.
        return list_of_connected_components

    # ==============================================================================
    # CONSTRAINTS EXPLORATION - CHECK COMPLETUDE OF CONSTRAINTS
    # ==============================================================================
    def check_completude_of_constraints(
        self,
        threshold: float = 1.0,
    ) -> bool:
        """
        The main method used to check if all possible constraints are known (not necessarily annotated because of the transitivity).
        The transitivity is taken into account, and the `threshold` parameters is used if constraints values are used in the constraints transitivity.

        Args:
            threshold (float, optional): The threshold used to define the transitivity link. Defaults to `1.0`.

        Returns:
            bool: Return `True` if all constraints are known, `False` otherwise.
        """

        # For each data ID...
        for data_ID in self._constraints_transitivity.keys():

            # ... if some data IDs are not linked by transitivity to this `data_ID` with a `"MUST_LINK"` or `"CANNOT_LINK"` constraints...
            if (
                len(self._constraints_transitivity[data_ID]["MUST_LINK"].keys())
                + len(self._constraints_transitivity[data_ID]["CANNOT_LINK"].keys())
            ) != len(self._constraints_transitivity.keys()):

                # ... then return `False`.
                return False

        # Otherwise, return `True`.
        return True

    # ==============================================================================
    # CONSTRAINTS EXPLORATION - GET MIN AND MAX NUMBER OF CLUSTERS
    # ==============================================================================
    def get_min_and_max_number_of_clusters(
        self,
        threshold: float = 1.0,
    ) -> Tuple[int, int]:
        """
        The main method used to get determine, for a clustering model that would not violate any constraints, the range of the possible clusters number.
        Minimum number of cluster is estimated by the coloration of the `"CANNOT_LINK"` constraints graph.
        Maximum number of cluster is defined by the number of `"MUST_LINK"` connected components.
        The transitivity is taken into account, and the `threshold` parameters is used if constraints values are used in the constraints transitivity.

        Args:
            threshold (float, optional): The threshold used to define the transitivity link. Defaults to `1.0`.

        Returns:
            Tuple[int,int]: The minimum and the maximum possible clusters numbers (for a clustering model that would not violate any constraints).
        """

        # Get the `"MUST_LINK"` connected components.
        list_of_connected_components: List[List[str]] = self.get_connected_components()

        ###
        ### 1. Estimation of minimum clusters number.
        ###

        # Get connected component ids.
        list_of_connected_component_ids: List[str] = [component[0] for component in list_of_connected_components]

        # Keep only components that have more that one `"CANNOT_LINK"` constraints.
        list_of_linked_connected_components_ids: List[str] = [
            component_id
            for component_id in list_of_connected_component_ids
            if len(self._constraints_transitivity[component_id]["CANNOT_LINK"].keys()) > 1  # noqa: WPS507
        ]

        # Get the `"CANNOT_LINK"` constraints.
        list_of_cannot_link_constraints: List[Tuple[int, int]] = [
            (i1, i2)
            for i1, data_ID1 in enumerate(list_of_linked_connected_components_ids)
            for i2, data_ID2 in enumerate(list_of_linked_connected_components_ids)
            if (i1 < i2)
            and (  # To get the complement, get all possible link that are not a `"CANNOT_LINK"`.
                data_ID2 in self._constraints_transitivity[data_ID1]["CANNOT_LINK"].keys()
            )
        ]

        # Create a networkx graph.
        cannot_link_graph: nx.Graph = nx.Graph()
        cannot_link_graph.add_nodes_from(list_of_connected_component_ids)  # Add components id as nodes in the graph.
        cannot_link_graph.add_edges_from(
            list_of_cannot_link_constraints
        )  # Add cannot link constraints as edges in the graph.

        # Estimate the minimum clusters number by trying to colorate the `"CANNOT_LINK"` constraints graph.
        # The lower bound has to be greater than 2.
        estimation_of_minimum_clusters_number: int = max(
            2,
            1
            + min(
                max(nx.coloring.greedy_color(cannot_link_graph, strategy="largest_first").values()),
                max(nx.coloring.greedy_color(cannot_link_graph, strategy="smallest_last").values()),
                max(nx.coloring.greedy_color(cannot_link_graph, strategy="random_sequential").values()),
                max(nx.coloring.greedy_color(cannot_link_graph, strategy="random_sequential").values()),
                max(nx.coloring.greedy_color(cannot_link_graph, strategy="random_sequential").values()),
            ),
        )

        ###
        ### 2. Computation of maximum clusters number.
        ###

        # Determine the maximum clusters number with the number of `"MUST_LINK"` connected components.
        maximum_clusters_number: int = len(list_of_connected_components)

        # Return minimum and maximum.
        return (estimation_of_minimum_clusters_number, maximum_clusters_number)

    # ==============================================================================
    # CONSTRAINTS TRANSITIVITY MANAGEMENT - GENERATE CONSTRAINTS TRANSITIVITY GRAPH
    # ==============================================================================
    def _generate_constraints_transitivity(
        self,
    ) -> None:
        """
        Generate `self._constraints_transitivity`, a constraints dictionary that takes into account the transitivity of constraints.
        Suppose there is no inconsistency in `self._constraints_dictionary`.
        It uses `Dict[str, None]` to simulate ordered sets.
        """

        self._constraints_transitivity: Dict[str, Dict[str, Dict[str, None]]] = {
            data_ID: {
                "MUST_LINK": {data_ID: None},  # Initialize MUST_LINK clusters constraints.
                "CANNOT_LINK": {},  # Initialize CANNOT_LINK clusters constraints.
            }
            for data_ID in self._constraints_dictionary.keys()
        }

        for data_ID1 in self._constraints_transitivity.keys():
            for data_ID2 in self._constraints_transitivity.keys():

                # Optimization : `self._constraints_dictionary` is symetric, so one pass is enough.
                if data_ID1 > data_ID2:
                    continue

                # Get the constraint between `data_ID1` and `data_ID2`.
                constraint = self._constraints_dictionary[data_ID1][data_ID2]

                # Add the constraint transitivity if the constraint is not `None`.
                if constraint is not None:
                    self._add_constraint_transitivity(
                        data_ID1=data_ID1,
                        data_ID2=data_ID2,
                        constraint_type=constraint[0],
                    )

    # ==============================================================================
    # CONSTRAINTS TRANSITIVITY MANAGEMENT - ADD CONSTRAINT TRANSITIVITY
    # ==============================================================================
    def _add_constraint_transitivity(
        self,
        data_ID1: str,
        data_ID2: str,
        constraint_type: str,
    ) -> bool:
        """
        Add constraint transitivity in `self._constraints_transitivity` between `data_ID1` and `data_ID2` for constraint type `constraint_type`.
        Suppose there is no inconsistency in `self._constraints_dictionary`.

        Args:
            data_ID1 (str): The first data ID that is concerned for this constraint transitivity addition.
            data_ID2 (str): The second data ID that is concerned for this constraint transitivity addition.
            constraint_type (str): The type of the constraint to add. The type have to be `"MUST_LINK"` or `"CANNOT_LINK"`.

        Returns:
            bool: `True` when the transitivity addition is done.
        """

        ###
        ### Case 1 : `constraint_type` is `"MUST_LINK"`.
        ###
        if constraint_type == "MUST_LINK":

            # Define new common set of `"MUST_LINK"` data IDs,
            # by merging the sets of `"MUST_LINK"` data IDs for `data_ID1` and `data_ID2`.
            new_MUST_LINK_common_set: Dict[str, None] = {
                **self._constraints_transitivity[data_ID1]["MUST_LINK"],
                **self._constraints_transitivity[data_ID2]["MUST_LINK"],
            }

            # Define new common set of `"CANNOT_LINK"` data IDs,
            # by merging the sets of `"CANNOT_LINK"` data IDs for `data_ID1` and `data_ID2`.
            new_CANNOT_LINK_common_set: Dict[str, None] = {
                **self._constraints_transitivity[data_ID1]["CANNOT_LINK"],
                **self._constraints_transitivity[data_ID2]["CANNOT_LINK"],
            }

            # For each data that are now similar to `data_ID1` and `data_ID2`...
            for data_ID_ML in new_MUST_LINK_common_set.keys():
                # ... affect the new set of `"MUST_LINK"` constraints...
                self._constraints_transitivity[data_ID_ML]["MUST_LINK"] = new_MUST_LINK_common_set
                # ... and affect the new set of `"CANNOT_LINK"` constraints.
                self._constraints_transitivity[data_ID_ML]["CANNOT_LINK"] = new_CANNOT_LINK_common_set

            # For each data that are now different to `data_ID1` and `data_ID2`...
            for data_ID_CL in new_CANNOT_LINK_common_set.keys():
                # ... affect the new set of `"CANNOT_LINK"` constraints.
                self._constraints_transitivity[data_ID_CL]["CANNOT_LINK"] = {
                    **self._constraints_transitivity[data_ID_CL]["CANNOT_LINK"],
                    **new_MUST_LINK_common_set,
                }

        ###
        ### Case 2 : `constraint_type` is `"CANNOT_LINK"`.
        ###
        else:  # if constraint_type == "CANNOT_LINK":

            # Define new common set of `"CANNOT_LINK"` data IDs for data IDs that are similar to `data_ID1`.
            new_CANNOT_LINK_set_for_data_ID1: Dict[str, None] = {
                **self._constraints_transitivity[data_ID1]["CANNOT_LINK"],
                **self._constraints_transitivity[data_ID2]["MUST_LINK"],
            }

            # Define new common set of `"CANNOT_LINK"` data IDs for data IDs that are similar to `data_ID2`.
            new_CANNOT_LINK_set_for_data_ID2: Dict[str, None] = {
                **self._constraints_transitivity[data_ID2]["CANNOT_LINK"],
                **self._constraints_transitivity[data_ID1]["MUST_LINK"],
            }

            # For each data that are similar to `data_ID1`...
            for data_ID_like_data_ID1 in self._constraints_transitivity[data_ID1]["MUST_LINK"].keys():
                # ... affect the new list of `"CANNOT_LINK"` constraints.
                self._constraints_transitivity[data_ID_like_data_ID1]["CANNOT_LINK"] = new_CANNOT_LINK_set_for_data_ID1

            # For each data that are similar to `data_ID2`...
            for data_ID_like_data_ID2 in self._constraints_transitivity[data_ID2]["MUST_LINK"].keys():
                # ... affect the new list of `"CANNOT_LINK"` constraints.
                self._constraints_transitivity[data_ID_like_data_ID2]["CANNOT_LINK"] = new_CANNOT_LINK_set_for_data_ID2

        # Return `True`
        return True

    # ==============================================================================
    # CONSTRAINTS CONFLICT - GET INVOLVED DATA IDS IN A CONFLICT
    # ==============================================================================
    def get_list_of_involved_data_IDs_in_a_constraint_conflict(
        self,
        data_ID1: str,
        data_ID2: str,
        constraint_type: str,
    ) -> Optional[List[str]]:
        """
        Get all data IDs involved in a constraints conflict.

        Args:
            data_ID1 (str): The first data ID involved in the constraint_conflit.
            data_ID2 (str): The second data ID involved in the constraint_conflit.
            constraint_type (str): The constraint that create a conflict. The constraints can be `"MUST_LINK"` or `"CANNOT_LINK"`.

        Raises:
            ValueError: if `data_ID1`, `data_ID2`, `constraint_type` are not managed.

        Returns:
            Optional[List[str]]: The list of data IDs that are involved in the conflict. It matches data IDs from connected components of `data_ID1` and `data_ID2`.
        """

        # If `data_ID1` is not in the data IDs that are currently managed, then raises a `ValueError`.
        if data_ID1 not in self._constraints_dictionary.keys():
            raise ValueError("The `data_ID1` `'" + str(data_ID1) + "'` is not managed.")

        # If `data_ID2` is not in the data IDs that are currently managed, then raises a `ValueError`.
        if data_ID2 not in self._constraints_dictionary.keys():
            raise ValueError("The `data_ID2` `'" + str(data_ID2) + "'` is not managed.")

        # If the `constraint_conflict` is not in `self._allowed_constraint_types`, then raises a `ValueError`.
        if constraint_type not in self._allowed_constraint_types:
            raise ValueError(
                "The `constraint_type` `'"
                + str(constraint_type)
                + "'` is not managed. Allowed constraints types are : `"
                + str(self._allowed_constraint_types)
                + "`."
            )

        # Case of conflict (after trying to add a constraint different from the inferred constraint).
        if self.get_inferred_constraint(
            data_ID1, data_ID2
        ) is not None and constraint_type != self.get_inferred_constraint(data_ID1, data_ID2):
            return [
                data_ID
                for connected_component in self.get_connected_components()  # Get involved components.
                for data_ID in connected_component  # Get data IDs from these components.
                if data_ID1 in connected_component or data_ID2 in connected_component
            ]

        # Case of no conflict.
        return None



# -*- coding: utf-8 -*-

"""
* Name:         cognitivefactory.interactive_clustering.utils.preprocessing
* Description:  Utilities methods to apply NLP preprocessing.
* Author:       Erwan Schild
* Created:      17/03/2021
* Licence:      CeCILL (https://cecill.info/licences.fr.html)
"""

# ==============================================================================
# IMPORT PYTHON DEPENDENCIES
# ==============================================================================

import unicodedata  # To handle accents.
from typing import Dict  # To type Python code (mypy).

import spacy  # To apply spacy language models.

# from nltk.stem.snowball import SnowballStemmer  # To stemm texts.


# ==============================================================================
# NLP PREPROCESSING
# ==============================================================================
def preprocess(
    dict_of_texts: Dict[str, str],
    apply_stopwords_deletion: bool = False,
    apply_parsing_filter: bool = False,
    apply_lemmatization: bool = False,
    spacy_language_model: str = "fr_core_news_sm",
) -> Dict[str, str]:
    """
    A method used to preprocess texts.
    It applies simple preprocessing (lowercasing, punctuations deletion, accents replacement, whitespace deletion).
    Some options are available to delete stopwords, apply lemmatization, and delete tokens according to their depth in the denpendency tree.

    References:
        - _spaCy_: `Honnibal, M. et I. Montani (2017). spaCy 2 : Natural language understanding with Bloom embeddings, convolutional neural networks and incremental parsing.`
        - _spaCy_ language models: `https://spacy.io/usage/models`
        - _NLTK_: `Bird, Steven, Edward Loper and Ewan Klein (2009), Natural Language Processing with Python. O’Reilly Media Inc.`
        - _NLTK_ _'SnowballStemmer'_: `https://www.nltk.org/api/nltk.stem.html#module-nltk.stem.snowball`

    Args:
        dict_of_texts (Dict[str,str]): A dictionary that contains the texts to preprocess.
        apply_stopwords_deletion (bool, optional): The option to delete stopwords. Defaults to `False`.
        apply_parsing_filter (bool, optional): The option to filter tokens based on dependency parsing results. If set, it only keeps `"ROOT"` tokens and their direct children. Defaults to `False`.
        apply_lemmatization (bool, optional): The option to lemmatize tokens. Defaults to `False`.
        spacy_language_model (str, optional): The spaCy language model to use if vectorizer is spacy. The model has to be installed. Defaults to `"fr_core_news_sm"`.

    Raises:
        ValueError: Raises error if the `spacy_language_model` is not installed.

    Returns:
        Dict[str,str]: A dictionary that contains the preprocessed texts.

    Examples:
        ```python
        # Import.
        from cognitivefactory.interactive_clustering.utils.preprocessing import preprocess

        # Define data.
        dict_of_texts={
            "0": "Comment signaler une perte de carte de paiement ?",
            "1": "Quelle est la procédure pour chercher une carte de crédit avalée ?",
            "2": "Ma carte Visa a un plafond de paiment trop bas, puis-je l'augmenter ?",
        }

        # Apply preprocessing.
        dict_of_preprocessed_texts = preprocess(
            dict_of_texts=dict_of_texts
            apply_stopwords_deletion=True,
            apply_parsing_filter=False,
            apply_lemmatization=False,
            spacy_language_model="fr_core_news_sm",
        )

        # Print results.
        print("Expected results", ";", {"0": "signaler perte carte paiement", "1": "procedure chercher carte credit avalee", "2": "carte visa plafond paiment l augmenter",})
        print("Computed results", ":", dict_of_preprocessed_texts)
        ```
    """

    # Initialize dictionary of preprocessed texts.
    dict_of_preprocessed_texts: Dict[str, str] = {}

    # Initialize punctuation translator.
    punctuation_translator = str.maketrans(
        {
            punct: " "
            for punct in (
                ".",
                ",",
                ";",
                ":",
                "!",
                "¡",
                "?",
                "¿",
                "…",
                "•",
                "(",
                ")",
                "{",
                "}",
                "[",
                "]",
                "«",
                "»",
                "^",
                "`",
                "'",
                '"',
                "\\",
                "/",
                "|",
                "-",
                "_",
                "#",
                "&",
                "~",
                "@",
            )
        }
    )

    # Load vectorizer (spacy language model).
    try:
        spacy_nlp = spacy.load(
            name=spacy_language_model,
            disable=[
                # "tagger", # Needed for lemmatization.
                # "parser", # Needed for filtering on dependency parsing.
                "ner",  # Not needed
            ],
        )
    except OSError as err:  # `spacy_language_model` is not installed.
        raise ValueError("The `spacy_language_model` '" + str(spacy_language_model) + "' is not installed.") from err

    # Initialize stemmer.
    ####stemmer = SnowballStemmer(language="french")

    # For each text...
    for key, text in dict_of_texts.items():

        # Force string type.
        preprocessed_text: str = str(text)

        # Apply lowercasing.
        preprocessed_text = text.lower()

        # Apply punctuation deletion (before tokenization).
        preprocessed_text = preprocessed_text.translate(punctuation_translator)

        # Apply tokenization and spaCy pipeline.
        tokens = [
            token
            for token in spacy_nlp(preprocessed_text)
            if (
                # Spaces are not allowed.
                not token.is_space
            )
            and (
                # Punctuation are not allowed.
                not token.is_punct
                and not token.is_quote
            )
            and (
                # If set, stopwords are not allowed.
                (not apply_stopwords_deletion)
                or (not token.is_stop)
            )
            and (
                # If set, stopwords are not allowed.
                (not apply_parsing_filter)
                or (len(list(token.ancestors)) <= 1)
            )
        ]

        # Apply retokenization with lemmatization.
        if apply_lemmatization:

            preprocessed_text = " ".join([token.lemma_.strip() for token in tokens])

        # Apply retokenization without lemmatization.
        else:
            preprocessed_text = " ".join([token.text.strip() for token in tokens])

        # Apply accents deletion (after lemmatization).
        preprocessed_text = "".join(
            [char for char in unicodedata.normalize("NFKD", preprocessed_text) if not unicodedata.combining(char)]
        )

        # Store preprocessed text.
        dict_of_preprocessed_texts[key] = preprocessed_text

    return dict_of_preprocessed_texts
